#!/bin/bash
set -feu -o pipefail

PYTHON_EXECUTABLE=${PYTHON_EXECUTABLE:-"$(which python3.8)"}

if [ $# -lt 2 ]; then
    echo "Usage ./test install_folder hysop_folder [cache_dir] [backup_cache_dir]"
    exit 1
fi

if [ $# -gt 4 ]; then
    echo "Usage ./test install_folder hysop_folder [cache_dir] [backup_cache_dir]"
    exit 1
fi

if [ ! -d "$1" ]; then
    echo "Folder $1 does not exist."
    exit 1
fi

if [ ! -d "$2" ]; then
    echo "Folder $2 does not exist."
    exit 1
fi

INSTALL_DIR="$1"
HYSOP_DIR="$2"
HYSOP_CACHE_DIR="${HOME}/.cache"

if [ $# -gt 2 ]; then
    CACHE_DIR="$3"
    HAS_CACHE_DIR=true
else
    HAS_CACHE_DIR=false
fi

if [ $# -gt 3 ]; then
    BACKUP_CACHE_DIR="$4"
    HAS_BACKUP_CACHE_DIR=true
else
    HAS_BACKUP_CACHE_DIR=false
fi

if [ "${HAS_CACHE_DIR}" = true ]; then
    mkdir -p "${HYSOP_CACHE_DIR}"
    if [ -d "${CACHE_DIR}" ]; then
        echo "Cache directory '$CACHE_DIR' was found."
        rsync -rtu "${CACHE_DIR}/" "${HYSOP_CACHE_DIR}/"
    else
        # Untill gitlab allows cache on failure we need
        # to provide initial cache so that CI succeeds (< 1h tests)
        # See https://gitlab.com/gitlab-org/gitlab/-/issues/18969
        # Initial cache can be injected in the docker image, see hysop/ci/utils/run_ci.sh.
        echo "Cache directory '$CACHE_DIR' does not exist, trying to use backup cache directory."
        if [[ "${HAS_BACKUP_CACHE_DIR}" = true ]]; then
            if [[ -d "${BACKUP_CACHE_DIR}" ]]; then
                echo "Backup cache directory '${BACKUP_CACHE_DIR}' was found."
                rsync -rtu "${BACKUP_CACHE_DIR}/" "${HYSOP_CACHE_DIR}/"
            else
                echo "Backup directory '${BACKUP_CACHE_DIR}' does not exist."
            fi
        else
            echo "No backup cache directory has been specified."
        fi
    fi
    mkdir -p "${CACHE_DIR}"
fi

# Environment variables
export PYTHONPATH="${INSTALL_DIR}/lib/python3.8/site-packages:${INSTALL_DIR}"
export PYTHONHASHSEED=42  # get consistent hashes accross MPI processes
export PYOPENCL_COMPILER_OUTPUT=0
export MPLBACKEND='cairo'

# OpenMPI specific variables
export OMPI_ALLOW_RUN_AS_ROOT=1
export OMPI_ALLOW_RUN_AS_ROOT_CONFIRM=1
export OMPI_MCA_rmaps_base_oversubscribe=1 
export OMPI_MCA_btl_vader_single_copy_mechanism=none  # see https://github.com/open-mpi/ompi/issues/4948

# HySoP specific variables
export HYSOP_VERBOSE=0
export HYSOP_DEBUG=0
export HYSOP_PROFILE=0
export HYSOP_KERNEL_DEBUG=0
export HYSOP_TRACE_WARNINGS=1


echo "Trying to load hysop module:"
${PYTHON_EXECUTABLE} -c 'import hysop; print(hysop)'
echo "module import successful !"
echo

echo "Default testing OpenCL platform is:"
${PYTHON_EXECUTABLE} -c 'import hysop; from hysop.testsenv import iter_clenv; print(next(iter(iter_clenv())));'

RUN_TESTS=${RUN_TESTS:-true}
RUN_EXAMPLES=${RUN_EXAMPLES:-true}
RUN_LONG_TESTS=${RUN_LONG_TESTS:-false}

COMMON_TEST_OPTIONS=''
TEST_DIR="${HYSOP_DIR}"
COMMON_EXAMPLE_OPTIONS='-VNC -d16 -cp float -maxit 2 --autotuner-max-candidates 1 --save-checkpoint --checkpoint-dump-freq 0 --checkpoint-dump-period 0 --checkpoint-dump-last --checkpoint-dump-times'
EXAMPLE_DIR="${HYSOP_DIR}/../hysop_examples/examples"

hysop_test() {
     test=$1
     echo 
     echo "TESTING $1"
     echo "========$(printf '=%.0s' `seq ${#1}`)"
     ${PYTHON_EXECUTABLE} -Wd "${TEST_DIR}/${1}" ${@:2} ${COMMON_TEST_OPTIONS} 
     echo
}
example_test() {
     test=$1
     echo 
     echo "EXAMPLE $1"
     echo "========$(printf '=%.0s' `seq ${#1}`)"
     ${PYTHON_EXECUTABLE} -Wd "${EXAMPLE_DIR}/${1}" ${@:2} ${COMMON_EXAMPLE_OPTIONS}
     echo
}

if [ "$RUN_TESTS" = true ]; then
    hysop_test "core/arrays/tests/test_array.py"
    hysop_test "core/graph/tests/test_graph.py"
    hysop_test "fields/tests/test_fields.py"
    hysop_test "numerics/tests/test_fft.py"
    hysop_test "operator/tests/test_analytic.py"
    hysop_test "operator/tests/test_transpose.py"
    hysop_test "operator/tests/test_fd_derivative.py"
    hysop_test "operator/tests/test_absorption.py"
    hysop_test "operator/tests/test_penalization.py"
    hysop_test "operator/tests/test_velocity_correction.py"
    hysop_test "operator/tests/test_restriction_filter.py"
    hysop_test "operator/tests/test_scales_advection.py"
    hysop_test "operator/tests/test_bilevel_advection.py"
    hysop_test "operator/tests/test_directional_advection.py"
    hysop_test "operator/tests/test_directional_diffusion.py"
    hysop_test "operator/tests/test_directional_stretching.py"
    hysop_test "operator/tests/test_custom_symbolic.py"
    hysop_test "operator/tests/test_spectral_derivative.py"
    hysop_test "operator/tests/test_spectral_curl.py"
    hysop_test "operator/tests/test_diffusion.py"
    hysop_test "operator/tests/test_poisson.py"
    hysop_test "operator/tests/test_solenoidal_projection.py"
    hysop_test "operator/tests/test_poisson_curl.py"
    ${HYSOP_DIR}/fields/tests/test_cartesian.sh
    ${HYSOP_DIR}/core/tests/test_checkpoint.sh
fi

if [ "${RUN_LONG_TESTS}" = true ]; then
    hysop_test "backend/device/codegen/kernels/tests/test_directional_advection.py"
    hysop_test "backend/device/codegen/kernels/tests/test_directional_remesh.py"
fi

if [ "${RUN_EXAMPLES}" = true ]; then
    example_test "analytic/analytic.py"
    example_test "scalar_diffusion/scalar_diffusion.py"
    example_test "scalar_advection/scalar_advection.py"
    example_test "scalar_advection/levelset.py"
    example_test "multiresolution/scalar_advection.py"
    example_test "shear_layer/shear_layer.py"
    example_test "taylor_green/taylor_green.py" '-impl python'
    example_test "taylor_green/taylor_green.py" '-impl opencl'
    example_test "bubble/periodic_bubble.py"
    example_test "bubble/periodic_bubble_levelset.py"
    example_test "bubble/periodic_bubble_levelset_penalization.py" #LLVM bug for DP
    example_test "bubble/periodic_jet_levelset.py"
    example_test "particles_above_salt/particles_above_salt_periodic.py"
    example_test "particles_above_salt/particles_above_salt_symmetrized.py"
fi

if [ "${HAS_CACHE_DIR}" = true ]; then
    rsync -rtu "${HYSOP_CACHE_DIR}/" "${CACHE_DIR}/"
    find "${CACHE_DIR}" -name '*.lock' -delete
fi

exit 0
