
import contextlib, random
from contextlib import contextmanager

import operator, hashlib

from hysop import __VERBOSE__, __KERNEL_DEBUG__

from hysop.tools.misc import Utils
from hysop.constants import np

from hysop.backend.device.codegen.base.opencl_codegen import OpenClCodeGenerator
from hysop.backend.device.codegen.base.kernel_codegen import KernelCodeGenerator
from hysop.backend.device.codegen.base.variables      import CodegenVariable, CodegenVectorClBuiltin
from hysop.backend.device.opencl.opencl_types          import OpenClTypeGen
from hysop.backend.device.codegen.base.utils          import WriteOnceDict, ArgDict

from hysop.backend.device.opencl import cl, clCharacterize
from hysop.backend.device.opencl.opencl_env import OpenClEnvironment
from hysop.backend.device.opencl.opencl_kernel import OpenClKernelLauncher
from hysop.backend.device.kernel_autotuner import KernelAutotuner
from hysop.backend.device.kernel_autotuner_config import KernelAutotunerConfig

class BandwidthKernel(KernelCodeGenerator):

    @staticmethod
    def codegen_name(vtype,nreads,nwrites):
        vtype = vtype.replace(' ','_')
        name = 'bandwidth_{}_{}_{}'.format(vtype,nreads,nwrites)
        return name

    def __init__(self, typegen, vtype, nreads, nwrites, known_vars=None):

        assert nreads>0 and nwrites>0

        kernel_args = self.gen_kernel_arguments(typegen, vtype)

        name = BandwidthKernel.codegen_name(vtype, nreads, nwrites)

        super(BandwidthKernel,self).__init__(
                name=name,
                typegen=typegen,
                work_dim=1,
                kernel_args=kernel_args,
                known_vars=known_vars)

        self.vtype   = vtype
        self.nreads  = nreads
        self.nwrites = nwrites
        self.gencode()


    def gen_kernel_arguments(self, typegen, vtype):
        kargs = ArgDict()
        kargs['dst'] = CodegenVariable(ctype=vtype,name='dst',ptr=True,typegen=typegen,
                nl=True,ptr_restrict=True,ptr_volatile=True, storage='__global')
        kargs['src'] = CodegenVariable(ctype=vtype,name='src',ptr=True,typegen=typegen,
                nl=True,ptr_restrict=True,ptr_volatile=True,storage='__global',const=True)
        return kargs

    def gencode(self):
        s = self
        vtype = s.vtype

        global_id    = s.vars['global_id']
        global_size  = s.vars['global_size']

        dst   = s.vars['dst']
        src   = s.vars['src']

        buf = CodegenVariable(ctype=vtype,name='buffer',typegen=self.typegen,init=0)

        with s._kernel_():
            global_size.declare(s,const=True)
            global_id.declare(s,const=True)
            s.jumpline()

            buf.declare(s)
            for i in range(self.nreads):
                offset = '{}+{}uL*{}'.format(global_id(),i, global_size())
                code = '{} {}= {};'.format(buf(), '+*-/'[i%4], src[offset])
                s.append(code)

            for i in range(self.nwrites):
                offset = '{}+{}uL*{}'.format(global_id(),i,global_size())
                code = '{} = {}+{};'.format(dst[offset], buf(), i)
                s.append(code)

if __name__ == '__main__':
    from hysop.backend.device.codegen.base.test import _test_typegen

    tg = _test_typegen('float')
    vtype = 'float4'

    ck = BandwidthKernel(tg, vtype, nreads=10, nwrites=1,
            known_vars={'global_size':1024, 'local_size': 512})
    ck.edit()
