import sympy as sm

from hysop.tools.types import check_instance, first_not_None, to_tuple
from hysop.tools.decorators import debug
from hysop.fields.continuous_field import ScalarField, Field
from hysop.topology.cartesian_descriptor import CartesianTopologyDescriptors
from hysop.backend.device.opencl.operator.custom_symbolic import OpenClCustomSymbolicOperator
from hysop.symbolic.relational import Assignment

class OpenClAnalyticField(OpenClCustomSymbolicOperator):
    """
    Applies an analytic formula, given by user, on its field.
    Formula is given as one or more sympy expressions.
    """

    @debug
    def __new__(cls, field, formula, variables, **kwds):
        name = kwds.pop('name', 'analytic_{}'.format(field.name))
        return super(OpenClAnalyticField, cls).__new__(cls, name=None, exprs=None,
                variables=variables, **kwds)

    @debug
    def __init__(self, field, formula, variables, **kwds):
        """
        Initialize a Analytic operator on the python backend.

        Apply a user-defined formula onto a field, possibly
        dependent on space variables and external fields/parameters.

        Parameters
        ----------
        field: hysop.field.continuous_field.ScalarField
            Continuous field to be modified.
        formula : sm.Basic or array-like of sm.Basic
            field.nb_components symbolic expressions as a tuple.
        variables: dict
            Dictionary of fields as keys and topology descriptors as values.
        kwds: dict, optional
            Base class arguments.
        """
        formula = to_tuple(formula)
        check_instance(field, ScalarField)
        check_instance(formula, tuple, values=(type(None),sm.Basic), size=field.nb_components)
        check_instance(variables, dict, keys=Field, values=CartesianTopologyDescriptors)

        exprs = ()
        Fs = to_tuple(field.s())
        for (lhs,rhs) in zip(Fs, formula):
            if (rhs is None):
                continue
            exprs += (Assignment(lhs, rhs),)

        if not exprs:
            msg='All formulas are None. Give at least one expression.'
            raise ValueError(msg)

        name = kwds.pop('name', 'analytic_{}'.format(field.name))

        super(OpenClAnalyticField, self).__init__(name=name, exprs=exprs,
                variables=variables, **kwds)

    @classmethod
    def supports_mpi(cls):
        return True
