
import primefac, functools
from hysop.tools.numerics import float_to_complex_dtype
from hysop.tools.numpywrappers import npw
from hysop.tools.decorators  import debug
from hysop.tools.warning import HysopWarning
from hysop.operator.base.diffusion import DiffusionOperatorBase
from hysop.backend.device.opencl.opencl_symbolic import OpenClSymbolic
from hysop.core.graph.graph import op_apply
from hysop.core.memory.memory_request import MemoryRequest
from hysop.backend.device.opencl.opencl_fft import OpenClFFT
from hysop.backend.device.codegen.base.variables import dtype_to_ctype
from hysop.symbolic import local_indices_symbols
from hysop.symbolic.relational import LogicalAND, LogicalEQ, Assignment

class OpenClDiffusion(DiffusionOperatorBase, OpenClSymbolic):
    """
    Solves the diffusion equation using an OpenCL FFT backend.
    """

    @debug
    def __new__(cls, **kwds):
        return super(OpenClDiffusion, cls).__new__(cls, **kwds)

    @debug
    def __init__(self, **kwds):
        super(OpenClDiffusion, self).__init__(**kwds)
        queue = self.cl_env.default_queue

        nu, dt = self.nu.s, self.dt.s

        kernel_names = ()
        for (i,(Ft,Wn)) in enumerate(zip(self.forward_transforms, self.wave_numbers)):
            Fhs = Ft.output_symbolic_array('F{}_hat'.format(i))
            indices = local_indices_symbols[:Fhs.dim]

            kname = 'filter_diffusion_{}d_{}'.format(Fhs.dim, i)
            kernel_names += (kname,)

            F = 0
            for Wi in Wn:
                indexed_Wi = self.tg._indexed_wave_numbers[Wi]
                F += indexed_Wi
            expr = Assignment(Fhs, Fhs / (1 - nu*dt*F))

            self.require_symbolic_kernel(kname, expr)

        self._kernel_names = kernel_names

    @debug
    def setup(self, work):
        super(OpenClDiffusion, self).setup(work)
        queue = self.cl_env.default_queue

        diffusion_filters = ()
        for kname in self._kernel_names:
            kernel, up = self.symbolic_kernels[kname]
            kernel = lambda kernel=kernel, queue=queue, up=up: kernel(queue=queue, **up())
            diffusion_filters += (kernel,)

        eg = self.dFout.exchange_ghosts(build_launcher=True)
        if (eg is not None):
            eg = functools.partial(eg, queue=queue)

        self._diffusion_filters = diffusion_filters
        self._exchange_ghosts = eg

    @op_apply
    def apply(self, simulation, **kwds):
        """Solve the Diffusion equation."""
        super(OpenClDiffusion, self).apply(**kwds)
        exchange_ghosts = self._exchange_ghosts
        for (Ft,Bt,filter_diffusion) in zip(
                self.forward_transforms,
                self.backward_transforms,
                self._diffusion_filters):
            evt = Ft(simulation=simulation)
            evt = filter_diffusion()
            evt = Bt(simulation=simulation)
        if exchange_ghosts:
            evt = exchange_ghosts()
        self.plot(simulation=simulation)
