import cpuinfo
from psutil import virtual_memory

from hysop.constants                import default_order
from hysop.core.memory.allocator import AllocatorBase
from hysop.backend.host.host_buffer import HostBuffer

class HostAllocator(AllocatorBase):
    """
    Allocator that allocates HostBuffers
    """

    def __new__(cls, verbose=None):
        return super(HostAllocator, cls).__new__(cls, verbose=verbose)

    def __init__(self, verbose=None):
        super(HostAllocator, self).__init__(verbose=verbose)
        self.mem_size = virtual_memory().total

    def max_alloc_size(self):
        """Max allocatable size in bytes."""
        return self.mem_size

    def allocate(self, nbytes, **kwds):
        super(HostAllocator, self).allocate(nbytes=nbytes, **kwds)
        return HostBuffer(size=nbytes)

    def prefix(self):
        return '{}: '.format(self.full_tag)

    def is_on_host(self):
        """
        Return true if buffers are allocated in host memory.
        """
        return True

    def memory_pool(self, name, **kwds):
        """
        Construct a memory pool from this allocator.
        """
        from hysop.backend.host.host_mempool import MemoryPool, HostMemoryPool
        if isinstance(self, MemoryPool):
            msg='allocator is already a memory pool.'
            raise RuntimeError(msg)
        return HostMemoryPool(allocator=self, name=name, **kwds)

def __get_default_name():
    try:
        cpu_name = cpuinfo.cpuinfo.get_cpu_info()['brand']
        pos = cpu_name.find('@')
        if pos>0:
            cpu_name = cpu_name[:pos]
    except:
        cpu_name = 'CPU0'
    return cpu_name.strip()

default_host_allocator = HostAllocator()
default_host_mempool = default_host_allocator.memory_pool(name=__get_default_name())
