from hysop.constants import MemoryOrdering
from hysop.tools.decorators import debug, profile
from hysop.backend.host.host_operator import HostOperator
from hysop.operator.base.transpose_operator import TransposeOperatorBase
from hysop.core.graph.graph import op_apply

import numpy as np
try:
    import hptt
    HAS_HPTT=True
    # required version is: https://gitlab.com/keckj/hptt
except ImportError:
    HAS_HPTT=False
    import warnings
    from hysop.tools.warning import HysopPerformanceWarning
    msg='Failed to import HPTT module, falling back to slow numpy transpose. Required version is available at https://gitlab.com/keckj/hptt.'
    warnings.warn(msg, HysopPerformanceWarning)

class PythonTranspose(TransposeOperatorBase, HostOperator):
    """
    Inplace and out of place field transposition and permutations in general.
    Uses basic numpy transposition facility.
    """

    @debug
    def __new__(cls, **kwds):
        return super(PythonTranspose, cls).__new__(cls, **kwds)

    @debug
    def __init__(self, **kwds):
        """Initialize a Transpose operator on the python backend."""
        super(PythonTranspose, self).__init__(**kwds)


    def discretize(self):
        super(PythonTranspose, self).discretize()
        assert self.din.dtype == self.dout.dtype
        dtype = self.din.dtype
        if HAS_HPTT and (dtype in (np.float32, np.float64, np.complex64, np.complex128)):
            if self.is_inplace:
                self.exec_transpose = self.transpose_hptt_inplace
            else:
                self.exec_transpose = self.transpose_hptt_outofplace
        else:
            if self.is_inplace:
                self.exec_transpose = self.transpose_np_inplace
            else:
                self.exec_transpose = self.transpose_np_outofplace

    @debug
    def get_field_requirements(self):
        requirements = super(PythonTranspose, self).get_field_requirements()
        for (is_input, reqs) in requirements.iter_requirements():
            if (reqs is None):
                continue
            (field, td, req) = reqs
            req.memory_order = MemoryOrdering.ANY
        return requirements

    @classmethod
    def supports_mpi(cls):
        return True

    @op_apply
    def apply(self, **kwds):
        """ Transpose in or out of place."""
        super(PythonTranspose,self).apply(**kwds)
        self.exec_transpose(**kwds)

    def transpose_hptt_inplace(self, **kwds):
        axes = self.axes
        din, dout, dtmp = self.din, self.dout, self.dtmp.handle.view(np.ndarray)
        assert self.din.dfield is self.dout.dfield
        for i in range(din.nb_components):
            hptt.transpose(a=din.buffers[i], out=dtmp, axes=axes)
            dout.buffers[i][...] = dtmp

    def transpose_hptt_outofplace(self, **kwds):
        axes = self.axes
        din, dout = self.din, self.dout
        assert self.din.dfield is not self.dout.dfield
        for i in range(din.nb_components):
            hptt.transpose(a=din.buffers[i], out=dout.buffers[i], axes=axes)

    def transpose_np_inplace(self, **kwds):
        axes = self.axes
        din, dout, dtmp = self.din, self.dout, self.dtmp.handle.view(np.ndarray)
        assert self.din.dfield is self.dout.dfield
        for i in range(din.nb_components):
            dtmp[...] = np.transpose(din.buffers[i], axes=axes)
            dout.buffers[i][...] = dtmp

    def transpose_np_outofplace(self, **kwds):
        axes = self.axes
        din, dout = self.din, self.dout
        assert self.din.dfield is not self.dout.dfield
        for i in range(din.nb_components):
            dout.buffers[i][...] = np.transpose(din.buffers[i], axes=axes)

