from abc import ABCMeta, abstractmethod
from hysop import dprint
from hysop.tools.decorators import debug
from hysop.tools.types import first_not_None
from hysop.core.graph.computational_node import ComputationalGraphNode

class ComputationalGraphNodeGenerator(object, metaclass=ABCMeta):
    """
    A class that can generate multiple hysop.core.graph.ComputationalGraphNode.
    """

    @debug
    def __new__(cls, candidate_input_tensors, candidate_output_tensors,
            name=None, pretty_name=None, **kwds):
        return super(ComputationalGraphNodeGenerator, cls).__new__(cls, **kwds)

    @debug
    def __init__(self, candidate_input_tensors, candidate_output_tensors,
            name=None, pretty_name=None, **kwds):
        super(ComputationalGraphNodeGenerator, self).__init__(**kwds)
        candidate_input_tensors  = first_not_None(candidate_input_tensors, ())
        candidate_output_tensors = first_not_None(candidate_output_tensors, ())

        self.name = name or self.__class__.__name__
        self.pretty_name = first_not_None(pretty_name, self.name)
        self.nodes     = []
        self.generated = False
        self.candidate_input_tensors = set(filter(lambda x: x.is_tensor, candidate_input_tensors))
        self.candidate_output_tensors = set(filter(lambda x: x.is_tensor,candidate_output_tensors))

    @abstractmethod
    def _generate(self, **kargs):
        """
        Method that should return a list of generated nodes.
        """
        pass

    def _pre_generate(self, **kargs):
        """
        Optional method called before _generate.
        """
        pass

    def _post_generate(self, **kargs):
        """
        Optional method called after _generate.
        """
        pass

    @debug
    def generate(self, **kargs):
        """
        Recursively generate nodes.
        Do not override.
        """
        if not self.generated:
            self._pre_generate(**kargs)

            assert len(self.nodes)==0
            try:
                for op in self._generate(**kargs):
                    if (op is None):
                        continue
                    if isinstance(op, ComputationalGraphNode):
                        nodes = (op,)
                        self.candidate_input_tensors.update(op.input_tensor_fields)
                        self.candidate_output_tensors.update(op.output_tensor_fields)
                    elif isinstance(op,ComputationalGraphNodeGenerator):
                        nodes = op.generate(**kargs)
                        if (nodes is None):
                            msg='FATAL ERROR: {}::{}.generate() returned None, '
                            msg+='this is an implementation bug.'
                            msg=msg.format(type(op), op.name)
                            raise RuntimeError(msg)
                        self.candidate_input_tensors.update(op.candidate_input_tensors)
                        self.candidate_output_tensors.update(op.candidate_output_tensors)
                    else:
                        msg='Unknown node type {}.'.format(op.__class__)
                        raise TypeError(msg)
                    self.nodes += nodes
            except:
                msg='\nFailed to call generate() in class {}.\n'.format(self.__class__)
                print(msg)
                raise

            self.candidate_input_tensors  = set(filter(lambda x: x.is_tensor,
                self.candidate_input_tensors))
            self.candidate_output_tensors = set(filter(lambda x: x.is_tensor,
                self.candidate_output_tensors))

            self._post_generate(**kargs)
            self.generated = True
        return self.nodes

    def to_graph(self, name=None, **kwds):
        """Generates and convert an computational node generator to a graph."""
        msg='Generating graph from computational node generator {}.'.format(name)
        dprint(msg)

        from hysop.core.graph.computational_graph import ComputationalGraph
        name = name or '{}_graph'.format(self.name)
        graph = ComputationalGraph(name=name, **kwds)
        graph.push_nodes(self)
        return graph

    def build(self, name=None, outputs_are_inputs=False, **kwds):
        """Convert a computational node generator to a graph and prepares it for apply."""
        return self.to_graph(name=name).build(outputs_are_inputs=outputs_are_inputs, **kwds)

