
import traceback
from abc import ABCMeta, abstractmethod
from hysop import __VERBOSE__, __TRACE_MEMALLOCS__, __BACKTRACE_BIG_MEMALLOCS__
from hysop.tools.types import check_instance
from hysop.tools.units import bytes2str
from hysop.tools.types import first_not_None
from hysop.tools.handle import TaggedObject

class AllocatorBase(TaggedObject, metaclass=ABCMeta):
    """
    Base class for allocators.
    """

    is_deferred=False

    def __new__(cls, verbose, **kwds):
        return super(AllocatorBase, cls).__new__(cls, tag_prefix='al', **kwds)

    def __init__(self, verbose, **kwds):
        super(AllocatorBase, self).__init__(tag_prefix='al', **kwds)
        verbose = first_not_None(verbose, __TRACE_MEMALLOCS__)
        self._verbose = verbose

    def __eq__(self, other):
        return (self is other)
    def __ne__(self, other):
        return (self is not other)
    def __hash__(self):
        return id(self)

    def __call__(self, size, alignment=None):
        """
        Allocate nbytes aligned on min_alignment.
        If the first allocation fails, call the garbage collector and try again.
        It this fails a second time, raise a MemoryError.
        """

        try_count = 0
        while try_count < 2:
            try:
                if (alignment > 1):
                    buf = self.allocate_aligned(size=size, alignment=alignment)
                else:
                    buf = self.allocate(nbytes=size)
                if (buf is None):
                    msg='{}.allocate(): returned allocation is None.'.format(self.__class__)
                    raise ValueError(msg)
                return buf
            except MemoryError:
                try_count += 1
                if try_count == 2:
                    raise

            self.try_release_blocks()

    @abstractmethod
    def prefix(self):
        """Prefix for logs"""
        pass

    @abstractmethod
    def is_on_host(self):
        """
        Return true if buffers are allocated in host memory.
        """
        pass

    @abstractmethod
    def max_alloc_size(self):
        """Max allocatable size in bytes."""
        pass

    @abstractmethod
    def allocate(self, nbytes, verbose=True):
        """
        Allocate nbytes bytes of memory.
        No minimal memory alignment is guaranteed.
        If allocation fails, this method has to raise a MemoryError.
        """
        if (self._verbose or __BACKTRACE_BIG_MEMALLOCS__) and verbose:
            print('{}allocating block of size {}.'.format(
                    self.prefix(), bytes2str(nbytes)))
        if __BACKTRACE_BIG_MEMALLOCS__ and nbytes>64*1024*1024:
            print('[BIG ALLOCATION BACKTRACE]')
            print(''.join(traceback.format_stack()))
            print('[END OF TRACE]')
            print()

    def allocate_aligned(self, size, alignment):
        """
        Allocate size bytes aligned on alignment.
        """
        assert alignment>0
        assert (alignment & (alignment-1))==0, 'alignment is not a power of 2.'
        nbytes = size + alignment - 1
        if self._verbose:
            print('{}allocating block of size {}, to satisfy {} aligned on {} bytes.'.format(
                    self.prefix(), bytes2str(nbytes), bytes2str(size), alignment))
        return self.allocate(nbytes, verbose=False).aligned_view(alignment=alignment, size=size)

    def try_release_blocks(self):
        """
        Try to release memory blocks by calling the garbage collector.
        """
        import gc
        gc.collect()

    def free(self, buf):
        """
        Release the allocated buffer.
        """
        buf.release()

