"""Implementation for data transfer/redistribution between topologies

`.. currentmodule : hysop.core.mpi.redistribute

See hysop.operator.redistribute.Redistribute for automatic
redistribute deployment.

* :class:`~RedistributeOperatorBase` abstract base class
* :class:`~RedistributeIntra` for topologies/operators defined
  inside the same mpi communicator
* :class:`~RedistributeInter` for topologies/operators defined
  on two different mpi communicator
* :class:`~RedistributeOverlap` for topologies defined
  inside the same mpi parent communicator and
  with a different number of processes
"""

from hysop.constants import Backend, DirectionLabels
from hysop.tools.types import check_instance, to_set
from hysop.tools.decorators import debug, not_implemented
from hysop.tools.numpywrappers import npw, slices_empty
from hysop.tools.mpi_utils import get_mpi_order
from hysop.topology.cartesian_topology import Topology, CartesianTopology
from hysop.core.mpi.topo_tools import TopoTools
from hysop.core.mpi.bridge   import Bridge, BridgeOverlap, BridgeInter
from hysop.operator.base.redistribute_operator import RedistributeOperatorBase
from hysop.core.graph.graph import op_apply

DEBUG_REDISTRIBUTE=0

def _memcpy(dst, src, target_indices, source_indices):
    def _runtime_error():
        msg='Copy from {} to {} are not handled yet.'
        msg=msg.format(src.__class__, dst.__class__)
        raise RuntimeError(msg)

    assert (src.dtype == dst.dtype)
    skind = src.backend.kind
    tkind = dst.backend.kind

    evt = None
    if (skind==Backend.HOST):
        if (tkind == Backend.HOST):
            dst[target_indices] = src[source_indices]
        elif (tkind==Backend.OPENCL):
            from hysop.backend.device.opencl.opencl_copy_kernel_launchers import \
                                                                OpenClCopyBufferRectLauncher
            knl = OpenClCopyBufferRectLauncher.from_slices(
                varname='redistribute',
                src=src, dst=dst,
                src_slices=source_indices, dst_slices=target_indices)
            evt = knl(queue=dst.default_queue)
        else:
            _runtime_error()
    elif (skind==Backend.OPENCL):
        from hysop.backend.device.opencl.opencl_copy_kernel_launchers import \
                                                            OpenClCopyBufferRectLauncher
        if (tkind==Backend.HOST):
            knl = OpenClCopyBufferRectLauncher.from_slices(
                varname='redistribute',
                src=src, dst=dst,
                src_slices=source_indices, dst_slices=target_indices)
            evt = knl(queue=src.default_queue)
        elif (tkind==Backend.OPENCL):
            assert (src.backend.cl_env is dst.backend.cl_env)
            knl = OpenClCopyBufferRectLauncher.from_slices(
                varname='redistribute',
                src=src, dst=dst,
                src_slices=source_indices, dst_slices=target_indices)
            evt = knl(queue=src.default_queue)
        else:
            _runtime_error()
    else:
        _runtime_error()
    return evt


class RedistributeIntra(RedistributeOperatorBase):
    """Data transfer between two operators/topologies.
    Source and target must:
     *be CartesianTopology topologies with the same global resolution
     *be defined on the same communicator
     *work on the same number of mpi processes
    """

    @classmethod
    def can_redistribute(cls, source_topo, target_topo):
        tin  = source_topo
        tout = target_topo

        # source and target must be CartesianTopology topology defined on HostArrayBackend
        if not isinstance(source_topo, CartesianTopology):
            return False
        if not isinstance(target_topo, CartesianTopology):
            return False

        # source and target must have the same global resolution
        source_res = tin.mesh.grid_resolution
        target_res = tout.mesh.grid_resolution
        if not npw.array_equal(source_res, target_res):
            return False

        # defined on the same communicator
        # and work on the same number of mpi process
        if not TopoTools.compare_comm(tin.parent, tout.parent):
            return False

        return True

    def __new__(cls, **kwds):
        return super(RedistributeIntra, cls).__new__(cls, **kwds)

    def __init__(self, **kwds):
        """Data transfer between two operators/topologies defined on the
        same communicator

        Source and target must:
         *be defined on the same communicator
         *work on the same number of mpi process
         *work with the same global resolution
        """

        # Base class initialisation
        super(RedistributeIntra, self).__init__(**kwds)

        # Warning : comm from io_params will be used as
        # reference for all mpi communication of this operator.
        # --> rank computed in refcomm
        # --> source and target must work inside refcomm
        # If io_params is None, refcomm will COMM_WORLD.

    def discretize(self):
        super(RedistributeIntra,self).discretize()

        # we can create the bridge
        ifield = self.input_discrete_fields[self.variable]
        ofield = self.output_discrete_fields[self.variable]
        source_topo = ifield.topology
        target_topo = ofield.topology
        sstate = source_topo.topology_state
        tstate = target_topo.topology_state
        if (sstate.dim != tstate.dim) or (sstate.axes!=tstate.axes) \
                or (sstate.memory_order!=tstate.memory_order):
            msg='Topology state mismatch between source and target.'
            msg+='\nSource topology state:'
            msg+=str(sstate)
            msg+='\nTarget topology state:'
            msg+=str(tstate)
            raise RuntimeError(msg)

        assert all(source_topo.mesh.local_resolution == ifield.resolution)
        assert all(target_topo.mesh.local_resolution == ofield.resolution)

        self.bridge = Bridge(source_topo, target_topo, self.dtype, get_mpi_order(ifield.sdata))
        self._rank = self.bridge._rank

        # dictionnary which maps rank with mpi derived type
        # for send operations
        self._send = self.bridge.send_types()

        # dictionnay which maps rank with mpi derived type
        # for send operations
        self._receive = self.bridge.recv_types()

        self._has_requests = False

        self.dFin  = ifield
        self.dFout = ofield

    @op_apply
    def apply(self, **kwds):
        # Try different way to send vars?
        # - Buffered : copy all data into a buffer and send/recv
        # - Standard : one send/recv
        dFin, dFout = self.dFin, self.dFout

        super(RedistributeIntra,self).apply(**kwds)

        # --- Standard send/recv ---
        br = self.bridge

        # dictionnary which map rank/field name with a receive request
        self._r_request = {}
        # dictionnary which map rank/field name with a send request
        self._s_request = {}

        basetag = self.mpi_params.rank + 1

        # Comm used for send/receive operations
        # It must contains all proc. of source topo and
        # target topo.
        refcomm = self.bridge.comm

        v = self.variable
        local_evts = ()
        v_name = v.name

        # Deal with local copies of data
        if br.has_local_inter():
            dst = self._vtarget[v].sdata
            src = self._vsource[v].sdata
            axes = self._vtarget[v].topology_state.axes
            source_indices = br.local_source_ind()
            target_indices = br.local_target_ind()
            evt = _memcpy(dst, src,
                          target_indices,
                          source_indices)
            if (evt is not None):
                local_evts += (evt,)

        # Transfers to other mpi processes
        for rk in self._receive:
            if (rk == self._rank):
                continue
            assert (src.backend.kind is Backend.HOST) and (dst.backend.kind is Backend.HOST)
            recvtag = basetag * 989 + (rk + 1) * 99
            mpi_type = self._receive[rk]
            dst = self._vtarget[v].sdata
            self._r_request[v_name + str(rk)] = \
                refcomm.Irecv([dst.handle, 1, mpi_type], source=rk, tag=recvtag)
            self._has_requests = True
        for rk in self._send:
            if (rk == self._rank):
                continue
            assert (src.backend.kind is Backend.HOST) and (dst.backend.kind is Backend.HOST)
            sendtag = (rk + 1) * 989 + basetag * 99
            mpi_type = self._send[rk]
            src = self._vsource[v].sdata
            self._s_request[v_name + str(rk)] = \
                refcomm.Issend([src.handle, 1, mpi_type], dest=rk, tag=sendtag)
            self._has_requests = True

        for evt in local_evts:
            evt.wait()

        if self._has_requests:
            for rk in self._r_request:
                self._r_request[rk].Wait()
            for rk in self._s_request:
                self._s_request[rk].Wait()
        self._has_requests = False

        if DEBUG_REDISTRIBUTE:
            print('resolution, compute_resolution, ghosts, compute_slices')
            print(dFin.resolution, dFin.compute_resolution, dFin.ghosts, dFin.compute_slices)
            print(dFout.resolution, dFout.compute_resolution, dFout.ghosts, dFout.compute_slices)
            print()
            print('BEFORE')
            dFout.print_with_ghosts()

        dFout.exchange_ghosts()

        if DEBUG_REDISTRIBUTE:
            print('AFTER')
            dFout.print_with_ghosts()
            mean_in  = refcomm.allreduce(dFin.sdata[dFin.compute_slices].sum().get())   / float(refcomm.size)
            mean_out = refcomm.allreduce(dFout.sdata[dFout.compute_slices].sum().get()) / float(refcomm.size)
            assert npw.isclose(mean_in, mean_out), '{} != {}'.format(mean_in, mean_out)

