"""
Abstract interfaces for physical domains description.
* :class:`~hysop.domain.domain.Domain`
* :class:`~hysop.domain.domain.DomainView`
"""
import hashlib
import numpy as np
from abc import ABCMeta, abstractmethod

from hysop.constants import HYSOP_DEFAULT_TASK_ID, HYSOP_DIM
from hysop.core.mpi import main_comm, MPI
from hysop.tools.parameters import MPIParams
from hysop.tools.decorators import debug
from hysop.tools.handle import RegisteredObject, TaggedObjectView
from hysop.tools.types import check_instance
from hysop.tools.numpywrappers import npw
from hysop.symbolic.frame import SymbolicFrame


class DomainView(TaggedObjectView, metaclass=ABCMeta):
    """Abstract base class for views on domains. """

    __slots__ = ('_domain', '_topology_state')

    @debug
    def __init__(self, topology_state, domain=None, **kwds):
        super(DomainView, self).__init__(obj_view=domain, **kwds)

    @debug
    def __new__(cls, topology_state, domain=None, **kwds):
        """Create and initialize a DomainView."""
        from hysop.topology.topology import TopologyState
        check_instance(topology_state, TopologyState)
        check_instance(domain, Domain, allow_none=True)

        obj = super(DomainView, cls).__new__(cls, obj_view=domain, **kwds)

        domain = domain or obj
        check_instance(domain, Domain)

        obj._domain = domain
        obj._topology_state = topology_state
        return obj

    def _get_domain(self):
        """Return the domain on which the view is on."""
        return self._domain

    def _get_topology_state(self):
        """Return the topology state altering this domain view."""
        return self._topology_state

    def _get_dim(self):
        """Return the dimension of the domain."""
        return self._domain._dim

    def _get_parent_comm(self):
        """Return the parent communicator used to create this domain."""
        return self._domain._parent_comm

    def _get_parent_rank(self):
        """Return the rank of the process in the parent communicator."""
        return self._domain._parent_rank

    def _get_task_comm(self):
        """
        Return the communicator that owns the current process.
        This is the sub-communicator which has been obtained by splitting.
        the parent communicator by colors (proc_tasks).
        """
        return self._domain._task_comm

    def _get_task_rank(self):
        """Return the rank of the process in the task communicator."""
        return self._domain._task_rank

    def _get_machine_comm(self):
        """
        Return the communicator that owns the current process.
        This is the sub-communicator which has been obtained by splitting.
        the parent communicator by machine name.
        """
        return self._domain._machine_comm

    def _get_machine_rank(self):
        """Return the rank of the process in the machine communicator."""
        return self._domain._machine_rank

    def _get_proc_tasks(self):
        """Return mapping between mpi process rank and task identifier."""
        return self._domain._proc_tasks

    def _get_registered_topologies(self):
        """
        Return the dictionary of all topologies already built on this domain,
        with topology ids as keys and :class:`~hysop.topology.topology.Topology` as values.
        """
        return self._domain._registered_topologies

    def _get_frame(self):
        """Get symbolic frame associated to this domain."""
        return self._domain._frame

    def task_on_proc(self, parent_rank):
        """Get task identifier for a given mpi process (parent communicator rank)."""
        if parent_rank >= len(self._domain._proc_tasks):
            msg = 'Unknown rank {} in parent communicator.'.format(parent_rank)
            raise ValueError(msg)
        return self._domain._proc_tasks[parent_rank]

    def current_task(self):
        """Get task number of the current mpi process."""
        return self.task_on_proc(self._domain._parent_rank)

    def is_on_task(self, params):
        """Test if the current process corresponds to param task."""
        if isinstance(params, MPIParams):
            task_id = params.task_id
        elif isinstance(params, (int, npw.integer)):
            task_id = params
        else:
            msg = 'Could not extract task_id from type {}.'
            msg = msg.format(type(params))
            raise TypeError(msg)
        return (self.current_task() == task_id)

    def print_topologies(self):
        """Print all topologies registered on the domain."""
        print(self.short_description() + ' defined the following topologies:')
        for topo in self._domain._registered_topologies.values():
            print('  *'+topo.short_description())

    @abstractmethod
    def short_description(self):
        """Return a short description of this domain as a string."""
        pass

    @abstractmethod
    def long_description(self):
        """Return a long description of this domain as a string."""
        pass

    def __eq__(self, other):
        if not isinstance(other, DomainView):
            return NotImplemented
        eq = (self._domain is other._domain)
        eq &= (self._topology_state == other._topology_state)
        return eq

    def __ne__(self, other):
        if not isinstance(other, DomainView):
            return NotImplemented
        eq = (self._domain is other._domain)
        eq &= (self._topology_state == other._topology_state)
        return not eq

    def __hash__(self):
        return id(self._domain) ^ hash(self._topology_state)

    def __str__(self):
        """Equivalent to self.long_description()"""
        return self.long_description()

    domain = property(_get_domain)
    dim = property(_get_dim)
    proc_tasks = property(_get_proc_tasks)
    parent_comm = property(_get_parent_comm)
    parent_rank = property(_get_parent_rank)
    task_comm = property(_get_task_comm)
    task_rank = property(_get_task_rank)
    machine_comm = property(_get_machine_comm)
    machine_rank = property(_get_machine_rank)
    registered_topologies = property(_get_registered_topologies)
    frame = property(_get_frame)


class Domain(RegisteredObject, metaclass=ABCMeta):
    """Abstract base class for the description of physical domains. """

    @debug
    def __init__(self, dim, parent_comm=None, proc_tasks=None, **kwds):
        super(Domain, self).__init__(dim=dim, parent_comm=parent_comm,
                proc_tasks=proc_tasks, tag_prefix='d', **kwds)

    @debug
    def __new__(cls, dim, parent_comm=None, proc_tasks=None, **kwds):
        """
        Create or get an existing physical domain of given dim on a specified MPI
        communicator and specific tasks.

        Parameters
        ----------
        dim : integer, optional
            dim of the domain.
        parent_comm : MPI.Intracomm, optional
            Parent communicator which may be split.
            If not given this will be hysop.core.mpi.main_comm.
        proc_tasks : array like of ints, optional
            Mapping between mpi process rank and task identifier.
            If not given all procs will be on task HYSOP_DEFAULT_TASK_ID.

        Attributes
        ----------
        dim : int
            Dimension of the domain.
        proc_tasks : tuple of ints
            Mapping between mpi process rank and task identifier.
        parent_comm: MPI.Intracomm
            Return the parent communicator used to create this domain.
        parent_rank: int
            Return the rank of the process in the parent communicator.
        task_comm : MPI.IntraComm
            Return the communicator that owns the current process.
            This is the sub-communicator which has been obtained by splitting
            the parent communicator by colors (proc_tasks).
        task_rank: int
            Return the rank of the process in the task communicator.
        registered_topologies : dict
            Dictionary of all topologies already built on this domain
            with topology ids as keys and :class:`~hysop.topology.topology.Topology` as values.

        Notes
        -----
        *Parent communicator is split according to proc_tasks.
        *About MPI Tasks
            proc_tasks[n] = 12 means that task 12 owns proc n
                or equivalently that proc n is dedicated to task 12.
        *A dupped parent_comm will return another idenpendent domain instance,
          because MPI communicators are hashed trough their python object id.
        """

        dim = int(dim)
        parent_comm = parent_comm or main_comm
        proc_tasks = proc_tasks or [HYSOP_DEFAULT_TASK_ID, ] * parent_comm.Get_size()
        proc_tasks = npw.asdimarray(proc_tasks)
        assert proc_tasks.size == parent_comm.Get_size()

        npw.set_readonly(proc_tasks)

        # double check types, to be sure RegisteredObject will work as expected
        check_instance(dim, int)
        check_instance(parent_comm, MPI.Intracomm)
        check_instance(proc_tasks, npw.ndarray, dtype=HYSOP_DIM)

        obj = super(Domain, cls).__new__(cls, dim=dim, parent_comm=parent_comm,
                                         proc_tasks=proc_tasks, tag_prefix='d', **kwds)

        if not obj.obj_initialized:
            obj.__initialize(dim, parent_comm, proc_tasks)

        return obj

    @debug
    def __initialize(self, dim, parent_comm, proc_tasks):
        parent_rank = parent_comm.Get_rank()
        parent_size = parent_comm.Get_size()

        if len(set(proc_tasks)) == 1:
            task_comm = parent_comm.Dup()
        else:
            assert len(proc_tasks) == parent_size
            task_comm = parent_comm.Split(color=proc_tasks[parent_rank],
                                          key=parent_rank)

        task_rank = task_comm.Get_rank()

        # Build a per-machine communicator in order to get a rank on local machines
        # Split accoring to machine name hashed and converted to integer (strings generally differs only from a single character)
        machine_comm = parent_comm.Split(
            color=np.int32(int(hashlib.md5(MPI.Get_processor_name().encode('utf-8')).hexdigest(), 16) %
                           np.iinfo(np.int32).max),
            key=parent_rank)
        machine_rank = machine_comm.Get_rank()

        self._dim = dim
        self._parent_comm = parent_comm
        self._parent_rank = parent_rank
        self._task_comm = task_comm
        self._task_rank = task_rank
        self._machine_comm = machine_comm
        self._machine_rank = machine_rank
        self._proc_tasks = proc_tasks
        self._registered_topologies = {}
        self._frame = SymbolicFrame(dim=dim)

    def register_topology(self, topo):
        """Register a new topology on this domain.
        Do nothing if an equivalent topology is already
        in the list.
        """
        from hysop.topology.topology import Topology
        check_instance(topo, Topology)
        topo_id = topo.id
        if topo_id in self._registered_topologies:
            assert topo is self._registered_topologies[topo_id]
        else:
            self._registered_topologies[topo_id] = topo

    def remove_topology(self, topo):
        """
        Remove a topology from the list of this domain.
        Do nothing if the topology does not exist in the list.
        """
        from hysop.topology.topology import Topology
        check_instance(topo, Topology)
        topo_id = topo.id
        if topo_id in self._registered_topologies:
            self._registered_topologies.pop(topo_id)
        else:
            topo_id = -1
        return topo_id

    @abstractmethod
    def view(self, topology_state):
        """Return a view of this domain altered by some topology_state."""
        pass
