from abc import ABCMeta
import numpy as np

from hysop.tools.types       import check_instance, first_not_None, InstanceOf
from hysop.tools.decorators  import debug
from hysop.fields.continuous_field import Field
from hysop.parameters.tensor_parameter import TensorParameter
from hysop.topology.cartesian_descriptor import CartesianTopologyDescriptors
from hysop.core.memory.memory_request import MemoryRequest
from hysop.constants import ResidualError

class ConvergenceBase(object):
    """Common implementation interface for Convergence operator"""

    __default_method = {
        ResidualError: ResidualError.ABSOLUTE
    }

    __available_methods = {
        ResidualError: (InstanceOf(ResidualError), )
    }

    @classmethod
    def default_method(cls):
        dm = super(ConvergenceBase, cls).default_method()
        dm.update(cls.__default_method)
        return dm
    @classmethod
    def available_methods(cls):
        am = super(ConvergenceBase, cls).available_methods()
        am.update(cls.__available_methods)
        return am

    @debug
    def __new__(cls, variables, convergence=None,
                 implementation=None, **kwds):
        return super(ConvergenceBase, cls).__new__(cls,
                input_fields=None, output_params=None, **kwds)

    @debug
    def __init__(self, variables, convergence=None,
                 implementation=None, **kwds):
        """
        variables: dict
            dictionary of fields as keys and topologies as values.
        convergence: TensorParameter (optional)
            Parameter to store the components of ||u-u_old||_\infty or ||u-u_old||_\infty/||u||_\infty
        implementation: Implementation, optional, defaults to None
            target implementation, should be contained in available_implementations().
            If None, implementation will be set to default_implementation().
        kwds:
            Extra keywords arguments that will be passed towards implementation
            enstrophy operator __init__.
        """
        check_instance(convergence, TensorParameter, allow_none=True)
        check_instance(variables, dict, keys=Field, values=CartesianTopologyDescriptors)

        field = next(iter(variables))
        if convergence is None:
            convergence = TensorParameter(name="|residual|", dtype=field.dtype,
                                          shape=(field.nb_components, ),
                                          is_read_only=False, quiet=False,
                                          initial_value=np.asarray([np.finfo(field.dtype).max,]*field.nb_components))

        input_fields  = {field: variables[field]}
        output_params = {convergence.name: convergence}

        self.field = field
        self.convergence = convergence

        super(ConvergenceBase, self).__init__(input_fields=input_fields,
                                              output_params=output_params, **kwds)

    @debug
    def handle_method(self, method):
        super(ConvergenceBase, self).handle_method(method)
        self._residual_computation = method.pop(ResidualError)

    @debug
    def discretize(self):
        if self.discretized:
            return
        super(ConvergenceBase, self).discretize()
        self.dField = self.get_input_discrete_field(self.field)
