from hysop.tools.types import check_instance, first_not_None
from hysop.tools.decorators import debug
from hysop.fields.continuous_field import Field
from hysop.topology.cartesian_descriptor import CartesianTopologyDescriptors
from hysop.operator.base.poisson import PoissonOperatorBase
from hysop.symbolic.relational import Assignment
from hysop.symbolic.field import laplacian
from hysop.parameters.scalar_parameter import ScalarParameter


class DiffusionOperatorBase(PoissonOperatorBase):
    """
    Common base for spectral diffusion operator.
    """

    @debug
    def __new__(cls, Fin, Fout, variables,
                 nu, dt, name=None, pretty_name=None, **kwds):
        return super(DiffusionOperatorBase, cls).__new__(cls,
                Fin=Fin, Fout=Fout, variables=variables,
                name=name, pretty_name=pretty_name,
                input_params=None, **kwds)

    @debug
    def __init__(self, Fin, Fout, variables,
                       nu, dt,
                       name=None, pretty_name=None,
                       **kwds):
        """
        Diffusion operator base.

        Parameters
        ----------
        Fin : :class:`~hysop.fields.continuous_field.Field`
            The input field to be diffused.
        Fout:  :class:`~hysop.fields.continuous_field.Field`
            The output field to be diffused.
        variables: dictionary of fields:topology
            The choosed discretizations.
        nu: ScalarParameter.
            Diffusion coefficient.
        dt: ScalarParameter
            Timestep parameter that will be used for time integration.
        kargs:
            Base class parameters.

        Notes:
            *Equations:
                dF/dt = nu*Laplacian(F)
                in  = Win
                out = Wout

            *Implicit resolution in spectral space:
                F_hat(tn+1) = 1/(1-nu*dt*sum(Ki**2)) * F_hat(tn)
        """
        check_instance(nu, ScalarParameter)
        check_instance(dt, ScalarParameter)

        input_params   = { dt.name: dt,
                           nu.name: nu }

        default_name = 'Diffusion_{}_{}'.format(Fin.name, Fout.name)
        default_pretty_name = 'Diffusion_{}_{}'.format(Fin.pretty_name, Fout.pretty_name)
        name = first_not_None(name, default_name)
        pretty_name = first_not_None(name, default_pretty_name)

        super(DiffusionOperatorBase, self).__init__(
                Fin=Fin, Fout=Fout, variables=variables,
                name=name, pretty_name=pretty_name,
                input_params=input_params, **kwds)

        self.nu = nu
        self.dt = dt
