
from abc import ABCMeta, abstractmethod
from hysop.tools.decorators import debug, not_implemented
from hysop.tools.types import check_instance, to_set, first_not_None
from hysop.tools.sympy_utils import subscript
from hysop.constants import Backend
from hysop.core.graph.computational_operator import ComputationalGraphOperator
from hysop.topology.topology import Topology
from hysop.fields.continuous_field import ScalarField

class RedistributeOperatorBase(ComputationalGraphOperator, metaclass=ABCMeta):
    """
    Abstract interface to redistribute operators.
    """

    @classmethod
    @not_implemented
    def can_redistribute(cls, source_topo, target_topo):
        """
        Return true if this RedistributeOperatorBase can be applied
        to redistribute a variable from source_topo to target_topo,
        else return False.
        """
        pass

    @classmethod
    def supported_backends(cls):
        """
        return the backends that this operator's topologies can support.
        """
        return Backend.all

    def __new__(cls, variable, source_topo, target_topo,
                    name=None, pretty_name=None, **kwds):
        return super(RedistributeOperatorBase, cls).__new__(cls,
                name=name, pretty_name=pretty_name,
                input_fields=None, output_fields=None, **kwds)

    def __init__(self, variable, source_topo, target_topo,
                    name=None, pretty_name=None, **kwds):
        """
        Parameters
        ----------
        variable: :class:`~hysop.field.continuous.ScalarField`
            the variable to be distributed
        source_topo: :class:`~hysop.topology.topology.Topology`
            source mesh topology
        target_topo: :class:`~hysop.topology.topology.Topology`
            target mesh topology
        name: str, optional
            name of this operator
        pretty_name: str, optional
            pretty name of this operator
        """
        check_instance(variable, ScalarField)
        check_instance(source_topo, Topology)
        check_instance(target_topo, Topology)

        input_fields  = {variable: source_topo}
        output_fields = {variable: target_topo}

        default_name = 'R{},{}_{}'.format(source_topo.id, target_topo.id, variable.name)

        default_pname = 'R{}{}{}_{}'.format(
                                          subscript(source_topo.id),
                                          '→',
                                          subscript(target_topo.id),
                                          variable.pretty_name)

        pretty_name = first_not_None(pretty_name, name, default_pname)
        name        = first_not_None(name, default_name)

        super(RedistributeOperatorBase, self).__init__(
                name=name, pretty_name=pretty_name,
                input_fields=input_fields,
                output_fields=output_fields, **kwds)

        self.variable    = variable
        self.dtype       = variable.dtype

        self.source_topo = source_topo
        self.target_topo = target_topo

        # Set domain and mpi params
        self._set_domain_and_tasks()

    @debug
    def get_field_requirements(self):
        reqs = super(RedistributeOperatorBase,self).get_field_requirements()

        for field in self.input_fields:
            _, req = reqs.get_input_requirement(field)
            req.axes = None
            req.memory_order = None

        for field in self.output_fields:
            _, req = reqs.get_output_requirement(field)
            req.axes = None
            req.memory_order = None

        return reqs

    @debug
    def get_node_requirements(self):
        from hysop.core.graph.node_requirements import OperatorRequirements
        reqs = super(RedistributeOperatorBase, self).get_node_requirements()
        reqs.enforce_unique_topology_shape=False
        return reqs

    @debug
    def initialize(self, topgraph_method=None, **kwds):
        super(RedistributeOperatorBase,self).initialize(topgraph_method)

    @debug
    def discretize(self):
        super(RedistributeOperatorBase,self).discretize()
        # Dictionnary of discrete field to be sent and received
        v = self.variable
        self._vsource = {v: self.input_discrete_fields[v]}
        self._vtarget = {v: self.output_discrete_fields[v]}

    @classmethod
    def supports_multiple_field_topologies(cls):
        return True
    @classmethod
    def supports_multiple_topologies(cls):
        return True
    @classmethod
    def supports_mpi(cls):
        return True
    def get_preserved_input_fields(self):
        return set(self.input_fields.keys())


    def available_methods(self):
        return {}
    def default_method(self):
        return {}
    def handle_method(self, method):
        super(RedistributeOperatorBase,self).handle_method(method)

