from hysop.constants import Implementation
from hysop.core.graph.computational_node_frontend import ComputationalGraphNodeFrontend
from hysop.parameters.parameter import Parameter
from hysop.tools.types import check_instance
from hysop.fields.continuous_field import Field
from hysop.tools.decorators import debug
from hysop.topology.cartesian_descriptor import CartesianTopologyDescriptors


class CustomOperator(ComputationalGraphNodeFrontend):
    """
    Function should take parameters in the following order:
      1. all input fields
      2. all input parameters
      3. all output fields
      4. all output parameters
    Note that discrete fields are passed as arguments to the custom function.
    """

    @classmethod
    def implementations(cls):
        from hysop.backend.host.python.operator.custom import PythonCustomOperator

        __implementations = {
            Implementation.PYTHON: PythonCustomOperator,
        }
        return __implementations

    @classmethod
    def default_implementation(cls):
        return Implementation.PYTHON

    @debug
    def __new__(cls, func, invars=None, outvars=None, extra_args=None, ghosts=None, **kwds):
        return super(CustomOperator, cls).__new__(cls,
            func=func, invars=invars, outvars=outvars, extra_args=extra_args, ghosts=ghosts, **kwds)

    @debug
    def __init__(self, func, invars=None, outvars=None, extra_args=None, ghosts=None, **kwds):
        check_instance(invars, (tuple, list), values=(Field, Parameter),
                       allow_none=True)
        check_instance(outvars, (tuple, list), values=(Field, Parameter),
                       allow_none=True)
        check_instance(extra_args, tuple, allow_none=True)
        check_instance(ghosts, int, allow_none=True)
        from inspect import getargspec as signature  # should be inspect.signature in python 3
        nb_args = len(signature(func).args)
        nb_in_f, nb_in_p, nb_out_f, nb_out_p, nb_extra = 0, 0, 0, 0, 0
        if invars is not None:
            for v in invars:
                if isinstance(v, Field):
                    nb_in_f += v.nb_components
                elif isinstance(v, Parameter):
                    nb_in_p += 1
        if outvars is not None:
            for v in outvars:
                if isinstance(v, Field):
                    nb_out_f += v.nb_components
                elif isinstance(v, Parameter):
                    nb_out_p += 1
        if not extra_args is None:
            nb_extra = len(extra_args)
        msg = "function arguments ({}) did not match given in/out ".format(signature(func))
        msg += "fields and parameters ({} input fields, {} input params,".format(nb_in_f, nb_in_p)
        msg += " {} output fields, {} output params).".format(nb_out_f, nb_out_p)
        assert nb_args == nb_in_f + nb_in_p + nb_out_f + nb_out_p + nb_extra, msg

        super(CustomOperator, self).__init__(
            func=func, invars=invars, outvars=outvars, extra_args=extra_args, ghosts=ghosts, **kwds)

