"""
@file diffusion.py
Diffusion operator generator.
"""
from hysop.constants         import Implementation, HYSOP_REAL
from hysop.tools.types       import check_instance, first_not_None
from hysop.tools.enum        import EnumFactory
from hysop.tools.decorators  import debug
from hysop.fields.continuous_field import Field
from hysop.topology.cartesian_descriptor import CartesianTopologyDescriptors
from hysop.parameters.scalar_parameter import ScalarParameter

from hysop.operator.base.spectral_operator import SpectralComputationalGraphNodeFrontend
from hysop.backend.host.python.operator.diffusion   import PythonDiffusion
try:
    from hysop.backend.device.opencl.operator.diffusion import OpenClDiffusion
except ImportError:
    OpenClDiffusion = None
try:
    from hysop.backend.host.fortran.operator.diffusion  import DiffusionFFTW
except ImportError:
    DiffusionFFTW = None

class Diffusion(SpectralComputationalGraphNodeFrontend):
    """
    Interface the diffusion solver.
    Available implementations are: FORTRAN: FFTW based solver (legacy fortran)
                                   PYTHON:  generic python fft based solver (pyfftw, scipy or numpy)
                                   OPENCL:  generic opencl fft based solver (gpyfft)
    """

    __implementations = {
            Implementation.PYTHON:  PythonDiffusion,
            Implementation.OPENCL:  OpenClDiffusion,
            Implementation.FORTRAN: DiffusionFFTW,
    }

    @classmethod
    def implementations(cls):
        return cls.__implementations

    @classmethod
    def default_implementation(cls):
        return Implementation.PYTHON

    @debug
    def __new__(cls, Fin, variables, nu, dt,
            Fout=None, implementation=None,
            base_kwds=None, **kwds):
        base_kwds = first_not_None(base_kwds, {})
        return super(Diffusion, cls).__new__(cls,
                Fin=Fin, Fout=Fout,
                variables=variables, nu=nu, dt=dt,
                implementation=implementation,
                base_kwds=base_kwds, **kwds)

    @debug
    def __init__(self, Fin, variables, nu, dt,
            Fout=None, implementation=None,
            base_kwds=None, **kwds):
        """
        Initialize a Poisson operator frontend.
        Solves dF/dt = nu * Laplacian(F)

        Parameters
        ----------
        Fin: field
            input field that should be diffused
        Fout: field, optional, defaults to none
            output field that should be diffused.
            if None this will be set to Fin.
        variables: dict
            dictionary of fields as keys and topologies as values.
        nu: float or ScalarParameter
            Some implementations may only offer scalar nu.
        dt: ScalarParameter
            Timestep parameter that will be used for time integration.
        implementation: Implementation, optional, defaults to None
            target implementation, should be contained in available_implementations().
            If None, implementation will be set to default_implementation().
        base_kwds: dict, optional, defaults to None
            Base class keywords arguments.
            If None, an empty dict will be passed.
        kwds:
            Keywords arguments that will be passed towards implementation
            poisson operator __init__.

        Notes
        -----
        A diffusion operator implementation should at least support
        the hysop.operator.base.diffusion.DiffusionBase interface.
        """
        Fout = first_not_None(Fout, Fin)
        base_kwds = first_not_None(base_kwds, {})

        check_instance(Fin,  Field)
        check_instance(Fout, Field)
        check_instance(variables, dict, keys=Field, values=CartesianTopologyDescriptors)
        check_instance(base_kwds, dict, keys=str)
        check_instance(dt, ScalarParameter)
        check_instance(nu, (float,ScalarParameter))
        if not isinstance(nu, ScalarParameter):
            nu = ScalarParameter(name='nu', dtype=HYSOP_REAL,
                                 initial_value=nu, quiet=True)

        super(Diffusion, self).__init__(Fin=Fin, Fout=Fout,
                                        variables=variables, nu=nu, dt=dt,
                                        implementation=implementation,
                                        base_kwds=base_kwds, **kwds)

