
from hysop.constants import Implementation
from hysop.fields.continuous_field import Field, ScalarField
from hysop.parameters.tensor_parameter import TensorParameter
from hysop.parameters.scalar_parameter import ScalarParameter
from hysop.operator.min_max import MinMaxFieldStatisticsBase
from hysop.core.graph.computational_node_frontend import ComputationalGraphNodeFrontend
from hysop.topology.cartesian_descriptor import CartesianTopologyDescriptors
from hysop.tools.types import first_not_None, to_tuple, check_instance
from hysop.tools.sympy_utils import nabla
from hysop.tools.decorators import debug


class SpectralExternalForce(ComputationalGraphNodeFrontend):
    """
    Generate an operator to compute the curl of a symbolic expression.
    """

    @classmethod
    def implementations(cls):
        from hysop.backend.device.opencl.operator.external_force import OpenClSpectralExternalForce
        __implementations = {
                Implementation.OPENCL: OpenClSpectralExternalForce,
        }
        return __implementations

    @classmethod
    def default_implementation(cls):
        return Implementation.OPENCL

    @debug
    def __new__(cls, vorticity, Fext, dt, variables,
                    Fmin=None, Fmax=None, Finf=None,
                    all_quiet=False, pbasename=None, ppbasename=None,
                    implementation=None, base_kwds=None, **kwds):
        base_kwds = first_not_None(base_kwds, {})
        return super(SpectralExternalForce, cls).__new__(cls,
                vorticity=vorticity,
                Fext=Fext, dt=dt, variables=variables,
                Fmin=None, Fmax=None, Finf=None,
                candidate_input_tensors=None,
                candidate_output_tensors=None,
                implementation=implementation,
                base_kwds=base_kwds, **kwds)

    @debug
    def __init__(self, vorticity, Fext, dt, variables,
                    Fmin=None, Fmax=None, Finf=None,
                    all_quiet=False, pbasename=None, ppbasename=None,
                    implementation=None, base_kwds=None, **kwds):
        """
        Create an operator that computes the curl of a given input force field Fext.

        Only the following configurations are supported:
                     dim   nb_components  |   dim   nb_components
        vorticity:    2          1        |    3          3

        What is computed:
            tmp  = curl(Fext) by using a spectral backend
            Fmin = min(tmp)
            Fmax = max(tmp)
            Finf = max(abs(Fmin), abs(Fmax))
            W   += dt*tmp

        where Fext is computed from user given ExternalForce.

        Parameters
        ----------
        vorticity: hysop.field.continuous_field.Field
            Continuous field as input ScalarField or VectorField.
            All contained field have to live on the same domain.
        Fext: hysop.operator.external_force.ExternalForce
            Expression of the external force.
        dt: ScalarParameter
            Timestep paramater.
        F...: ScalarParameter, TensorParameter or boolean, optional
            TensorParameters should match the shape of tmp (see Notes).
            If set to True, the TensorParameter will be generated automatically.
        all_quiet: bool
            Force all autogenerated TensorParameter to be quiet.
            By default, only the autogenerated TensorParameters that are not required
            by the user are set to be quiet.
        pbasename: str, optional
            Parameters basename for created parameters.
            Defaults to 'curl_{}'.format(Fext.name).
        ppbasename: str, optional
            Parameters pretty basename for created parameters.
            Defaults to '|{} x {}|'.format(nabla, Fext.pretty_name).
        variables: dict
            dictionary of fields as keys and topologies as values.
        implementation: Implementation, optional, defaults to None
            target implementation, should be contained in available_implementations().
            If None, implementation will be set to default_implementation().
        kwds: dict, optional
            Extra parameters passed towards base class (MultiSpaceDerivatives).

        Notes
        -----
        If dim == 2, it is expected that:
            vorticity has only one component
            Fext has 2 components
            Expected parameters are ScalarParameters or TensorParameters of shape (1,)
        Else if dim == 3:
            vorticity has 3 components
            Fext has 3 components
            Expected parameters are TensorParameter of shape (3,)
        """
        from hysop.operator.base.external_force import ExternalForce
        base_kwds = first_not_None(base_kwds, {})
        check_instance(vorticity, Field)
        check_instance(Fext, ExternalForce)
        check_instance(dt, ScalarParameter)
        check_instance(variables, dict, keys=Field, values=CartesianTopologyDescriptors)

        # Pregenerate parameters so that we can directly store them in self.
        default_pbasename  = 'curl_{}'.format(Fext.name)
        default_ppbasename = '{}x{}'.format(nabla, Fext.pretty_name)
        pbasename  = first_not_None(pbasename,  default_pbasename)
        ppbasename = first_not_None(ppbasename, default_ppbasename)
        parameters = MinMaxFieldStatisticsBase.build_parameters(field=vorticity,
                components=None, dtype=None, all_quiet=all_quiet,
                Fmin=Fmin, Fmax=Fmax, Finf=Finf,
                pbasename=pbasename, ppbasename=ppbasename)

        (Fmin, Fmax, Finf) = tuple(parameters[k] for k in ('Fmin', 'Fmax', 'Finf'))

        check_instance(Fmin, TensorParameter, allow_none=True)
        check_instance(Fmax, TensorParameter, allow_none=True)
        check_instance(Finf, TensorParameter, allow_none=True)

        super(SpectralExternalForce, self).__init__(vorticity=vorticity,
                Fext=Fext, dt=dt, variables=variables,
                Fmin=Fmin, Fmax=Fmax, Finf=Finf,
                candidate_input_tensors=(vorticity,),
                candidate_output_tensors=(vorticity,),
                implementation=implementation,
                base_kwds=base_kwds, **kwds)

        self.Fmin = Fmin
        self.Fmax = Fmax
        self.Finf = Finf
