import sys

from hysop.testsenv import __ENABLE_LONG_TESTS__
from hysop.testsenv import iter_clenv
from hysop.tools.numpywrappers import npw
from hysop.tools.contexts import printoptions
from hysop.tools.types import first_not_None, to_tuple
from hysop.tools.io_utils import IO
from hysop.core.graph.computational_graph import ComputationalGraph
from hysop.operator.directional.advection_dir import DirectionalAdvection
from hysop.operator.misc import ForceTopologyState
from hysop.operator.advection import Advection
from hysop.parameters.scalar_parameter import ScalarParameter

from hysop import Field, Box
from hysop.methods import Remesh, TimeIntegrator
from hysop.constants import Implementation, DirectionLabels, Backend, HYSOP_REAL
from hysop.numerics.splitting.strang import StrangSplitting, StrangOrder
from hysop.numerics.odesolvers.runge_kutta import Euler, RK2, RK4
from hysop.numerics.remesh.remesh import RemeshKernel


class TestScalesAdvectionOperator(object):

    @classmethod
    def setup_class(cls,
                    enable_extra_tests=__ENABLE_LONG_TESTS__,
                    enable_debug_mode=False):

        IO.set_default_path('/tmp/hysop_tests/test_scales_advection')

        if enable_debug_mode:
            cls.size_min = 8
            cls.size_max = 9
        else:
            cls.size_min = 10
            cls.size_max = 23

        cls.enable_extra_tests = enable_extra_tests
        cls.enable_debug_mode = enable_debug_mode

    @classmethod
    def teardown_class(cls):
        pass

    def _test(self, dim, is_inplace,
              size_min=None, size_max=None):
        assert dim > 0

        # periodic boundaries removes one computational point
        # so we add one here.
        size_min = first_not_None(size_min, self.size_min) + 1
        size_max = first_not_None(size_max, self.size_max) + 1

        shape = tuple(npw.random.randint(low=size_min, high=size_max, size=dim).tolist())

        if self.enable_extra_tests:
            flt_types = (HYSOP_REAL, )
            time_integrators = (RK2, )
            remesh_kernels = (Remesh.L2_1, Remesh.L4_2, Remesh.L6_4)
            velocity_cfls = (0.62, 1.89)
        else:
            flt_types = (HYSOP_REAL,)
            time_integrators = (RK2,)
            remesh_kernels = (Remesh.L4_2,)
            velocity_cfls = (1.89,)

        domain = Box(length=(2*npw.pi,)*dim)
        for dtype in flt_types:
            Vin = Field(domain=domain, name='Vin', dtype=dtype,
                        nb_components=dim)
            Sin = Field(domain=domain, name='Sin', dtype=dtype,
                        nb_components=5)
            Sout = Field(domain=domain, name='Sout', dtype=dtype,
                         nb_components=5)
            for time_integrator in time_integrators:
                for remesh_kernel in remesh_kernels:
                    for velocity_cfl in velocity_cfls:
                        print
                        self._test_one(time_integrator=time_integrator, remesh_kernel=remesh_kernel,
                                       shape=shape, dim=dim, dtype=dtype,
                                       is_inplace=is_inplace, domain=domain,
                                       Vin=Vin, Sin=Sin, Sout=Sout, velocity_cfl=velocity_cfl)

    @classmethod
    def __velocity_init(cls, data, coords, component, axes):
        if component in axes:
            data[...] = +1.0
        else:
            data[...] = 0.0

    @classmethod
    def __scalar_init(cls, data, coords, component, offsets=None):
        offsets = first_not_None(offsets, (0.0,)*len(coords))
        assert len(coords) == len(offsets)
        data[...] = 1.0/(component+1)
        for (c, o) in zip(coords, offsets):
            data[...] *= npw.cos(c+o)

    def _test_one(self, time_integrator, remesh_kernel,
                  shape, dim,
                  dtype, is_inplace, domain, velocity_cfl,
                  Vin, Sin, Sout):

        print('\nTesting {}D ScalesAdvection_{}_{}: inplace={} dtype={} shape={}, cfl={}'.format(
            dim, time_integrator.name(), remesh_kernel,
            is_inplace, dtype.__name__, shape, velocity_cfl), end=' ')
        if is_inplace:
            vin = Vin
            sin, sout = Sin, Sin
            variables = {vin: shape, sin: shape}
        else:
            vin = Vin
            sin, sout = Sin, Sout
            variables = {vin: shape, sin: shape, sout: shape}

        # Use optimal timestep, ||Vi||_inf is 1 on a per-axis basis
        dt = ScalarParameter('dt', initial_value=npw.nan)
        dt.value = (0.99 * velocity_cfl) / (max(shape)-1)

        implementations = [Implementation.FORTRAN, ]
        ref_impl = implementations[0]

        method = {TimeIntegrator: time_integrator, Remesh: remesh_kernel}

        def iter_impl(impl):
            base_kwds = dict(velocity=vin, advected_fields=sin, advected_fields_out=sout, dt=dt,
                             variables=variables, implementation=impl,
                             method=method, name='advection_{}'.format(str(impl).lower()))
            force_tstate = ForceTopologyState(fields=variables.keys(),
                                              variables=variables,
                                              backend=Backend.HOST)
            graph = ComputationalGraph(name='test_graph')
            if impl is Implementation.PYTHON:
                da = DirectionalAdvection(velocity_cfl=velocity_cfl, **base_kwds)
                split = StrangSplitting(splitting_dim=dim,
                                        order=StrangOrder.STRANG_SECOND_ORDER)
                split.push_operators(da)
                graph.push_nodes(split, force_tstate)
                yield 'default', graph
            elif impl is Implementation.FORTRAN:
                assert dim == 3, "Scales is only 3D"
                adv = Advection(**base_kwds)
                graph.push_nodes(adv, force_tstate)
                yield 'SCALES', graph
            else:
                msg = 'Unknown implementation to test {}.'.format(impl)
                raise NotImplementedError(msg)

        # Compare to other implementations
        advec_axes = (tuple(),)
        advec_axes += tuple((x,) for x in range(dim))
        if (dim > 1):
            advec_axes += (tuple(range(dim)),)

        reference_fields = {}
        for impl in implementations:
            print('\n >Implementation {}:'.format(impl), end=' ')
            is_ref = (impl == ref_impl)
            for (sop, graph) in iter_impl(impl):
                print('\n   *{}: '.format(sop), end=' ')

                graph.build()
                for axes in advec_axes:
                    ref_outputs = reference_fields.setdefault(axes, {})
                    napplies = 10
                    Vi = npw.asarray([+1.0 if (i in axes) else +0.0
                                      for i in range(dim)], dtype=dtype)

                    dvin = graph.get_input_discrete_field(vin).as_contiguous_dfield()
                    dsin = graph.get_input_discrete_field(sin).as_contiguous_dfield()
                    dsout = graph.get_output_discrete_field(sout).as_contiguous_dfield()
                    dsref = dsout.clone()

                    sys.stdout.write('.')
                    sys.stdout.flush()
                    try:
                        dvin.initialize(self.__velocity_init, axes=axes)
                        dsin.initialize(self.__scalar_init)
                        dsout.initialize(self.__scalar_init)

                        _input = tuple(dsin.data[i].get().handle.copy()
                                       for i in range(dsin.nb_components))
                        S0 = dsin.integrate()

                        for k in range(napplies+1):
                            if (k > 0):
                                graph.apply()

                            output = tuple(dsout.data[i].get().handle.copy()[dsout.compute_slices]
                                           for i in range(dsout.nb_components))

                            for i in range(dsout.nb_components):
                                mask = npw.isfinite(output[i][dsout.compute_slices])
                                if not mask.all():
                                    msg = '\nFATAL ERROR: Output is not finite on axis {}.\n'.format(
                                        i)
                                    print(msg)
                                    npw.fancy_print(output[i], replace_values={
                                                    (lambda a: npw.isfinite(a)): '.'})
                                    raise RuntimeError(msg)

                            if is_ref:
                                dxk = -Vi*(k)*dt()
                                dsref.initialize(self.__scalar_init, offsets=dxk.tolist())
                                d = dsout.distance(dsref, p=2)
                                if npw.any(d > 1e-1):
                                    print('FATAL ERROR: Could not match analytic advection.')
                                    print('DSOUT')
                                    print(dsout.sdata[dsout.compute_slices])
                                    print('DSREF')
                                    print(dsref.sdata[dsref.compute_slices])
                                    print('DSREF - DSOUT')
                                    print(dsout.sdata[dsout.compute_slices].get() - dsref.sdata[dsref.compute_slices].get())
                                    msg = 'Test failed with V={}, k={}, dxk={}, inter-field L2 distances are {}.'
                                    msg = msg.format(Vi, k, to_tuple(
                                        dxk, cast=float), to_tuple(d, cast=float))
                                    raise RuntimeError(msg)
                                ref_outputs[k] = output
                            else:
                                assert k in ref_outputs
                                reference = ref_outputs[k]
                                for i in range(dsout.nb_components):
                                    di = npw.abs(reference[i] - output[i])
                                    max_di = npw.max(di)
                                    neps = 10000
                                    max_tol = neps*npw.finfo(dsout.dtype).eps
                                    if (max_di > max_tol):
                                        print('FATAL ERROR: Could not match other implementation results.')
                                        print('\nComparisson failed at step {} and component {}:'.format(k, i))
                                        for (j, dv) in dvin.iter_fields():
                                            print('VELOCITY INPUT {}'.format(DirectionLabels[j]))
                                            print(dv.sdata[dv.compute_slices])
                                        print('SCALAR INPUT')
                                        print(_input[i])
                                        print('SCALAR REFERENCE')
                                        print(reference[i])
                                        print('SCALAR OUTPUT')
                                        print(output[i])
                                        print('ABS(REF - OUT)')
                                        npw.fancy_print(di, replace_values={
                                                        (lambda a: a < max_tol): '.'})
                                        print()
                                        msg = 'Output did not match reference output for component {} at time step {}.'
                                        msg += '\n > max computed distance was {}.'.format(max_di)
                                        msg += '\n > max tolerence was set to {} ({} eps).'.format(
                                            max_tol, neps)
                                        msg = msg.format(i, k)
                                        raise RuntimeError(msg)
                            Si = dsout.integrate()
                            if not npw.all(npw.isfinite(Si)):
                                msg = 'Integral is not finite. Got {}.'.format(Si)
                                raise RuntimeError(msg)
                            if (npw.abs(Si-S0) > 1e-3).any():
                                msg = 'Scalar was not conserved on iteration {}, expected {} but got {}.'
                                msg = msg.format(k,
                                                 to_tuple(S0, cast=float),
                                                 to_tuple(Si, cast=float))
                                raise RuntimeError(msg)
                    except:
                        sys.stdout.write('\bx\n\n')
                        sys.stdout.flush()
                        raise

    def perform_tests(self):
        # Scales is only 3D
        self._test(dim=3, is_inplace=True)
        print()

    def test_3D(self):
        self._test(dim=3, is_inplace=True)


if __name__ == '__main__':
    import hysop
    TestScalesAdvectionOperator.setup_class(enable_extra_tests=False,
                                            enable_debug_mode=False)

    test = TestScalesAdvectionOperator()

    with printoptions(threshold=10000, linewidth=240,
                      nanstr='nan', infstr='inf',
                      formatter={'float': lambda x: '{:>6.2f}'.format(x)}):
        test.perform_tests()

    TestScalesAdvectionOperator.teardown_class()
