import sys, datetime

from hysop.constants import Backend, MemoryOrdering
from hysop.tools.types import check_instance, first_not_None, to_tuple, to_list
from hysop.tools.string_utils import vprint_banner
from hysop.tools.contexts import Timer
from hysop.tools.decorators import debug
from hysop.tools.parameters import MPIParams
from hysop.core.checkpoints import CheckpointHandler
from hysop.core.graph.computational_graph import ComputationalGraph
from hysop.tools.string_utils import vprint_banner, vprint


class Problem(ComputationalGraph):

    def __init__(self, name=None, method=None, mpi_params=None,
                 check_unique_clenv=True, **kwds):
        mpi_params = first_not_None(mpi_params, MPIParams()) # enforce mpi params for problems
        super(Problem, self).__init__(name=name, method=method, mpi_params=mpi_params, **kwds)
        self._do_check_unique_clenv = check_unique_clenv

    @debug
    def insert(self, *ops):
        self.push_nodes(*ops)
        return self

    @debug
    def build(self, args=None, allow_subbuffers=False, outputs_are_inputs=True):
        with Timer() as tm:
            msg = self.build_problem(args=args,
                                     allow_subbuffers=allow_subbuffers,
                                     outputs_are_inputs=outputs_are_inputs)
            if msg:
                msg = ' Problem {} achieved, exiting ! '.format(msg)
                vprint_banner(msg, at_border=2)
                sys.exit(0)
        size = self.mpi_params.size
        avg_time = self.mpi_params.comm.allreduce(tm.interval) / size
        msg = ' Problem building took {} ({}s)'
        if size > 1:
            msg += ', averaged over {} ranks. '.format(size)
        msg = msg.format(datetime.timedelta(seconds=round(avg_time)),
                         avg_time)
        vprint_banner(msg, spacing=True, at_border=2)

        if (args is not None) and args.stop_at_build:
            msg = ' Problem has been built, exiting. '
            vprint_banner(msg, at_border=2)
            sys.exit(0)

    def build_problem(self, args, allow_subbuffers, outputs_are_inputs=True):
        if (args is not None) and args.stop_at_initialization:
            return 'initialization'
        vprint('\nInitializing problem...')
        # Initialize and Discretize first the other Problems
        for node in [_ for _ in self.nodes if isinstance(_, Problem)]:
            node.initialize(outputs_are_inputs=True, topgraph_method=None)
            node.discretize()
        self.initialize(outputs_are_inputs=outputs_are_inputs, topgraph_method=None)

        if (args is not None) and args.stop_at_discretization:
            return 'discretization'
        vprint('\nDiscretizing problem...')
        self.discretize()

        if (args is not None) and args.stop_at_work_properties:
            return 'work properties retrieval'
        vprint('\nGetting work properties...')
        work = self.get_work_properties()

        if (args is not None) and args.stop_at_work_allocation:
            return 'work allocation'
        vprint('\nAllocating work...')
        work.allocate(allow_subbuffers=allow_subbuffers)

        if (args is not None) and args.stop_at_setup:
            return 'setup'
        vprint('\nSetting up problem...')
        self.setup(work)

    def discretize(self):
        super(Problem, self).discretize()
        if self._do_check_unique_clenv:
            self.check_unique_clenv()

    def check_unique_clenv(self):
        cl_env, first_op = None, None
        for op in self.nodes:
            for topo in set(op.input_fields.values()).union(op.output_fields.values()):
                if (topo.backend.kind == Backend.OPENCL):
                    if (cl_env is None):
                        first_op = op
                        cl_env = topo.backend.cl_env
                    elif (topo.backend.cl_env is not cl_env):
                        msg = ''
                        msg += '\nOpenCl environment mismatch between operator {} and operator {}.'
                        msg = msg.format(first_op.name, op.name)
                        msg += '\n{}'.format(cl_env)
                        msg += '\n and'
                        msg += '\n{}'.format(topo.backend.cl_env)
                        msg += '\n If this is required, override check_unique_clenv().'
                        raise RuntimeError(msg)

    def initialize_field(self, field, **kwds):
        """Initialize a field on all its input and output topologies."""
        initialized = set()
        for op in self.nodes:
            # give priority to tensor field initialization
            for op_fields in (self.input_discrete_tensor_fields, self.output_discrete_tensor_fields,
                              op.input_discrete_tensor_fields, op.output_discrete_tensor_fields,
                              op.input_discrete_fields, op.output_discrete_fields):
                if (field in op_fields):
                    dfield = op_fields[field]
                    if all((df in initialized) for df in dfield.discrete_fields()):
                        # all contained scalar fields were already initialized
                        continue
                    else:
                        components = ()
                        for (component, scalar_dfield) in dfield.nd_iter():
                            if (scalar_dfield._dfield not in initialized):
                                components += (component,)
                        dfield.initialize(components=components, **kwds)
                        initialized.update(dfield.discrete_fields())
        if not initialized:
            msg = 'FATAL ERROR: Could not initialize field {}.'.format(field.name)
            raise RuntimeError(msg)

    @debug
    def solve(self, simu, dry_run=False, dbg=None,
              report_freq=10, plot_freq=10,
              checkpoint_handler=None, **kwds):

        if dry_run:
            vprint()
            vprint_banner('** Dry-run requested, skipping simulation. **')
            return

        simu.initialize()

        check_instance(checkpoint_handler, CheckpointHandler, allow_none=True)
        checkpoint_handler.create_checkpoint_template(self, simu)
        checkpoint_handler.load_checkpoint(self, simu)

        vprint('\nSolving problem...')
        with Timer() as tm:
            while not simu.is_over:
                vprint()
                simu.print_state()
                self.apply(simulation=simu, dbg=dbg, **kwds)
                should_dump_checkpoint = checkpoint_handler.should_dump(simu) # determined before simu advance
                simu.advance(dbg=dbg, plot_freq=plot_freq)
                if should_dump_checkpoint:
                    checkpoint_handler.save_checkpoint(self, simu)
                if report_freq and (simu.current_iteration % report_freq) == 0:
                    self.profiler_report()

        size = self.mpi_params.size
        avg_time = self.mpi_params.comm.allreduce(tm.interval) / size
        msg = ' Simulation took {} ({}s)'
        if size > 1:
            msg += ', averaged over {} ranks. '.format(size)
        msg += '\n  for {} iterations ({}s per iteration) '
        msg = msg.format(datetime.timedelta(seconds=round(avg_time)),
                         avg_time, max(simu.current_iteration+1, 1),
                         avg_time / max(simu.current_iteration+1, 1))
        vprint_banner(msg, spacing=True, at_border=2)

        simu.finalize()
        checkpoint_handler.finalize(self.mpi_params)
        self.final_report()

        if (dbg is not None):
            dbg('final iteration', nostack=True)

    def final_report(self):
        self.profiler_report()

    @debug
    def finalize(self):
        vprint('Finalizing problem...')
        super(Problem, self).finalize()

