import sympy as sm

from hysop.tools.numpywrappers import npw
from hysop.tools.types import check_instance, first_not_None, to_tuple
from hysop.symbolic import Expr, Symbol

InstructionTermination = ''

class Select(Expr):
    def __new__(cls, a, b, c, *args):
        """ Equivalent to ternary operator: (c ? b : a) """
        obj = super(Select, cls).__new__(cls, a, b, c, *args)
        obj.a = a
        obj.b = b
        obj.c = c
        return obj

    def __str__(self):
        return '({} ? {} : {})'.format(self.c, self.b, self.a)
    def __repr__(self):
        return 'Select({}, {}, {})'.format(repr(self.a), repr(self.b), repr(self.c))
    def _sympystr(self, printer):
        return '({} ? {} : {})'.format(printer._print(self.c),
                                       printer._print(self.b),
                                       printer._print(self.a))


class Expand(Expr):
    """Expand a vector by a given factor.
       v=v.xy => v.xxyy (factor = 2)
    """
    def __new__(cls, expr, factor, *args):
        obj = super(Expand, cls).__new__(cls, expr, factor, *args)
        obj.expr = expr
        obj.factor = factor
        return obj
    def __str__(self):
        return 'Expand({}, {})'.format(self.expr, self.factor)
    def __repr__(self):
        return self.__str__()
    def _sympystr(self, printer):
        return 'Expand({}, {})'.format(printer._print(self.expr), self.factor)


class BroadCast(Expr):
    """
    BroadCast a vector by a given factor.
     v=v.xy => v.xyxy (factor = 2)
    """
    def __new__(cls, expr, factor):
        obj = super(BroadCast, cls).__new__(cls, expr, factor)
        obj.expr = expr
        obj.factor = factor
        return obj
    def __str__(self):
        return 'BroadCast({}, {})'.format(self.expr, self.factor)
    def __repr__(self):
        return self.__str__()
    def _sympystr(self, printer):
        return 'BroadCast({}, {})'.format(printer._print(self.expr), self.factor)


class Cast(Expr):
    """
    Cast a scalar or a vector to another type.
    """
    def __new__(cls, expr, dtype):
        obj = super(Cast, cls).__new__(cls, expr, dtype)
        obj.expr = expr
        obj.dtype = dtype
        return obj
    def __str__(self):
        return 'Cast({}, {})'.format(self.expr, self.dtype)
    def __repr__(self):
        return self.__str__()
    def _sympystr(self, printer):
        return 'Cast({}, {})'.format(printer._print(self.expr), self.dtype)


class CodeSection(Expr):
    def __new__(cls, *exprs):
        assert len(exprs)>0
        assert all(type(e) not in (tuple, list, set, frozenset, dict) for e in exprs)
        obj = super(CodeSection, cls).__new__(cls, *exprs)
        return obj

    def _sympystr(self, printer):
        return 'CodeSection([{}])'.format(
                '; '.join(printer._print(a) for a in self.args))

    def _ccode(self, printer):
        codegen = printer.codegen
        with codegen._block_():
            for e in self.args:
                printer._print(e)
        return InstructionTermination

class MutexOp(Expr):
    def __new__(cls, mutexes, mutex_id, *args):
        obj = super(MutexOp, cls).__new__(cls, mutexes, mutex_id, *args)
        obj.mutexes  = mutexes
        obj.mutex_id = mutex_id
        return obj

class MutexLock(MutexOp):
    def _sympystr(self, printer):
        return 'MutexLock({}, {})'.format(
                printer._print(self.mutexes),
                printer._print(self.mutex_id))
    def _ccode(self, printer):
        codegen = printer.codegen
        spin='while (atomic_cmpxchg({}+{}, 0, 1) == 1);'
        spin=spin.format(printer._print(self.mutexes),
                         printer._print(self.mutex_id))
        codegen.append(spin)
        return InstructionTermination

class MutexUnlock(MutexOp):
    def _sympystr(self, printer):
        return 'MutexUnlock({}, {})'.format(
                printer._print(self.mutexes),
                printer._print(self.mutex_id))
    def _ccode(self, printer):
        codegen = printer.codegen
        spin='{}[{}] = 0;'
        spin=spin.format(printer._print(self.mutexes),
                         printer._print(self.mutex_id))
        codegen.append(spin)
        return InstructionTermination

def CriticalCodeSection(exprs, mutexes, mutex_id, *args):
        from hysop.symbolic.relational import Assignment
        exprs = to_tuple(exprs)
        mutex_lock   = MutexLock(mutexes, mutex_id)
        mutex_unlock = MutexUnlock(mutexes, mutex_id)
        exprs = (mutex_lock,) + exprs + (mutex_unlock,)
        return CodeSection(*exprs)


class ApplyStencil(Expr):
    def __new__(cls, expr, stencil):
        obj = super(ApplyStencil, cls).__new__(cls, expr, stencil)
        obj.expr    = expr
        obj.stencil = stencil
        return obj

    def _sympystr(self, printer):
        return 'ApplyStencil({}, {})'.format(printer._print(self.expr), self.stencil.coeffs)


class TimeIntegrate(Expr):
    """
    Represents variable time integration for code generation.

    Parameters
    ----------
    time_integrator: TimeIntegrator
    lhs : Field
    rhs : Expr
    """
    def __new__(cls, time_integrator, lhs, rhs):
        obj = super(TimeIntegrate, cls).__new__(cls, time_integrator, lhs, rhs)
        obj.time_integrator = time_integrator
        obj.lhs = lhs
        obj.rhs = rhs
        return obj

    def __str__(self):
        return '{} = {}({})'.format(self.lhs, self.time_integrator.name(), self.rhs)

    def _ccode(self):
        raise RuntimeError()

