
try:
   import cPickle as pickle
except:
   import pickle

import gzip, portalocker, contextlib, os, errno, uuid

machine_id = None
for path in ('/var/lib/dbus/machine-id', '/var/lib/yum/uuid'):
    try:
        with open(path) as f:
            machine_id = f.read().replace('\n', '')
    except:
        pass
    if (machine_id is not None):
        break
if (machine_id in  (None,'')):
    machine_id = uuid.getnode()

@contextlib.contextmanager
def lock_file(filepath, mode, compressed=True,
        timeout=3600, check_interval=1):
    """
    Opens a locked file with specified mode, possibly compressed.
    """
    _dir = os.path.dirname(filepath)

    try:
        if not os.path.isdir(_dir):
            try:
                os.makedirs(_dir)
            except OSError as e:
                if (e.errno != errno.EEXIST):
                    raise
        if not os.path.exists(filepath):
            open(filepath, 'a').close()
        with portalocker.Lock(filename=filepath, timeout=timeout, mode=mode,
                check_interval=check_interval) as fl:
            if compressed:
                with gzip.GzipFile(fileobj=fl, mode=mode) as f:
                    yield f
            else:
                yield fl
    except portalocker.exceptions.LockException as e:
        msg='\nFATAL ERROR: Could not obtain lock for file \'{}\' after waiting for {}s.\n'
        msg=msg.format(filepath, timeout)
        print(msg)
        raise e

@contextlib.contextmanager
def read_only_lock(filepath, compressed=True,
        timeout=3600, check_interval=1):
    """Opens a locked read only file, possibly compressed."""
    with lock_file(filepath=filepath, mode='rb', compressed=compressed,
            timeout=timeout, check_interval=check_interval) as f:
        yield f

@contextlib.contextmanager
def write_only_lock(filepath, compressed=True,
        timeout=3600, check_interval=1):
    """Opens a locked write only file, possibly compressed."""
    with lock_file(filepath=filepath, mode='wb', compressed=compressed,
            timeout=timeout, check_interval=check_interval) as f:
        yield f


def load_cache(filepath, match_type=dict, on_fail={}, **kwds):
    """Load pickled data from filepath atomically."""
    data = on_fail
    with read_only_lock(filepath, **kwds) as f:
        try:
            data = pickle.load(f)
            if not isinstance(data, match_type):
                raise pickle.UnpicklingError
        except (IOError, EOFError, pickle.UnpicklingError, AttributeError, TypeError):
            data = on_fail
    return data

def update_cache(filepath, key, data, match_type=dict, on_fail={}, **kwds):
    """
    Update cache entry in given file atomically with a (key,data) pair.
    Cached data is a pickled dictionnary.
    """
    cached_data = load_cache(filepath=filepath, match_type=match_type,
                                on_fail=on_fail, **kwds)
    cached_data[key] = data
    with write_only_lock(filepath=filepath, **kwds) as f:
        pickle.dump(cached_data, f)

def load_data_from_cache(filepath, key, match_type=dict, on_fail={}, **kwds):
    """Load cached data from a given file atomically with given key."""
    data = load_cache(filepath=filepath, match_type=match_type, on_fail=on_fail)
    if (key in data):
        return data[key]
    else:
        return None

def load_attributes_from_cache(filepath, key, instance, attrs, **kwds):
    """
    Load cached entries from a given file atomically.

    Cached data is assumed to be a dictionnary.
    If key is present in pickled data, try to get all
    given attributes data by keys given in attrs.
    Set instance attributes with those values.

    If one attribute is missing or key is not present
    in loaded data, set all values to None in instance
    and return False.

    Return True on success.
    """
    success = False
    data = load_cache(filepath=filepath, match_type=dict, **kwds)
    if (key in data):
        success = True
        data = data[key]
        for attr in attrs:
            if (attr not in data.keys()):
                for attr in attrs:
                    setattr(instance, attr, None)
                success = False
                break
            setattr(instance, attr, data[attr])
    return success
