from hysop.tools.types import first_not_None, to_tuple
from hysop.tools.sympy_utils import nabla, partial, subscript, subscripts, \
                                    exponent, exponents, xsymbol, get_derivative_variables

from sympy.printing.str import StrPrinter, StrReprPrinter
from sympy.printing.ccode import C99CodePrinter
from sympy.printing.latex import LatexPrinter

class BasePrinter(object):
    def print_Derivative(self, expr):
        (bvar, pvar, vvar, lvar) = print_all_names(expr.args[0])
        pvar = pvar
        all_xvars = get_derivative_variables(expr)
        xvars   = tuple(set(all_xvars))
        varpows = tuple(all_xvars.count(x) for x in xvars)
        bxvars  = tuple(print_name(x) for x in xvars)
        pxvars  = tuple(print_pretty_name(x) for x in xvars)
        vxvars  = tuple(print_var_name(x) for x in xvars)
        lxvars  = tuple(print_latex_name(x) for x in xvars)
        return DifferentialStringFormatter.format_pd(bvar, pvar, vvar, lvar,
                                                     bxvars, pxvars, vxvars, lxvars,
                                                     varpows=varpows)

    def _print(self, expr, **kwds):
        try:
            return super(BasePrinter, self)._print(expr, **kwds)
        except:
            print
            msg='FATAL ERROR: {} failed to print expression {}.'
            msg=msg.format(type(self).__name__, expr)
            print(msg)
            print
            raise


class NamePrinter(BasePrinter, StrReprPrinter):
    def _print(self, expr, **kwds):
        if hasattr(expr, 'name'):
            return expr.name
        elif hasattr(expr, '_name'):
            return expr._name
        return super(NamePrinter, self)._print(expr, **kwds)
    def _print_Derivative(self, expr):
        return super(NamePrinter, self).print_Derivative(expr)[0]
    def _print_Add(self, expr):
        return super(NamePrinter, self)._print_Add(expr).replace(' ', '')
    def _print_Mul(self, expr):
        return super(NamePrinter, self)._print_Mul(expr).replace(' ', '')
    def emptyPrinter(self, expr):
        msg='\n{} does not implement _print_{}(self, expr).'
        msg+='\nExpression is {}.'.format(expr)
        msg+='\nExpression type MRO is:'
        msg+='\n  *'+'\n  *'.join(t.__name__ for t in type(expr).__mro__)
        msg=msg.format(self.__class__.__name__, expr.__class__.__name__)
        raise NotImplementedError(msg)


class PrettyNamePrinter(BasePrinter, StrPrinter):
    def _print(self, expr, **kwds):
        if hasattr(expr, 'pretty_name'):
            return expr.pretty_name
        elif hasattr(expr, '_pretty_name'):
            return expr._pretty_name
        return super(PrettyNamePrinter, self)._print(expr, **kwds)
    def _print_Derivative(self, expr):
        return super(PrettyNamePrinter, self).print_Derivative(expr)[1]
    def emptyPrinter(self, expr):
        msg='\n{} does not implement _print_{}(self, expr).'
        msg+='\nExpression is {}.'.format(expr)
        msg+='\nExpression type MRO is:'
        msg+='\n  *'+'\n  *'.join(t.__name__ for t in type(expr).__mro__)
        msg=msg.format(self.__class__.__name__, expr.__class__.__name__)
        raise NotImplementedError(msg)


class VarNamePrinter(BasePrinter, C99CodePrinter):
    def _print(self, expr, **kwds):
        if hasattr(expr, 'var_name'):
            return expr.var_name
        elif hasattr(expr, '_var_name'):
            return expr._var_name
        return super(VarNamePrinter, self)._print(expr, **kwds).replace(' ', '')
    def _print_Derivative(self, expr):
        return super(VarNamePrinter, self).print_Derivative(expr)[2]
    def _print_Add(self, expr):
        s = super(VarNamePrinter, self)._print_Add(expr)
        s = s.replace(' + ', '_plus_').replace(' - ', '_minus_')
        s = s.replace('+', 'plus_').replace('-', 'minus_')
        return s
    def _print_Mul(self, expr):
        s = super(VarNamePrinter, self)._print_Mul(expr)
        s = s.replace(' * ', '_times_').replace('+', 'plus_').replace('-', 'minus_')
        return s
    def emptyPrinter(self, expr):
        msg='\n{} does not implement _print_{}(self, expr).'
        msg+='\nExpression is {}.'.format(expr)
        msg+='\nExpression type MRO is:'
        msg+='\n  *'+'\n  *'.join(t.__name__ for t in type(expr).__mro__)
        msg=msg.format(self.__class__.__name__, expr.__class__.__name__)
        raise NotImplementedError(msg)


class LatexNamePrinter(BasePrinter, LatexPrinter):
    def _print(self, expr, **kwds):
        if hasattr(expr, 'latex_name'):
            return expr.latex_name
        elif hasattr(expr, '_latex_name'):
            return expr._latex_name
        return super(LatexNamePrinter, self)._print(expr, **kwds)
    def _print_Derivative(self, expr):
        return super(LatexNamePrinter, self).print_Derivative(expr)[3]
    def _print_int(self, expr):
        return str(expr)
    def emptyPrinter(self, expr):
        msg='\n{} does not implement _print_{}(self, expr).'
        msg+='\nExpression is {}.'.format(expr)
        msg+='\nExpression type MRO is:'
        msg+='\n  *'+'\n  *'.join(t.__name__ for t in type(expr).__mro__)
        msg=msg.format(self.__class__.__name__, expr.__class__.__name__)
        raise NotImplementedError(msg)

pbn = NamePrinter()
ppn = PrettyNamePrinter()
#pvn = VarNamePrinter()
pln = LatexNamePrinter()

def print_name(expr):
    return pbn.doprint(expr)

def print_pretty_name(expr):
    return ppn.doprint(expr)

def print_var_name(expr):
    return VarNamePrinter().doprint(expr)

def print_latex_name(expr):
    return pln.doprint(expr)

def print_all_names(expr):
    name = print_name(expr)
    pretty_name = print_pretty_name(expr)
    var_name = print_var_name(expr)
    latex_name = print_latex_name(expr)
    return (name, pretty_name, var_name, latex_name)


def to_str(*args):
    if len(args)==1:
        args=to_tuple(args[0])
    def _to_str(x):
        return str(x)
    return tuple(_to_str(y) for y in args)

# exponents formatting functions
bexp_fn = lambda x: '^{}'.format(x) if (x>1) else ''
pexp_fn = lambda x, sep=',': exponents(x, sep=sep) if (x>1) else ''
vexp_fn = lambda x: 'e{}'.format(x) if (x>1) else ''
lexp_fn = lambda x: '^<LBRACKET>{}<RBRACKET>'.format(x) if (x>1) else ''

# powers formatting functions
bpow_fn = lambda x: '**{}'.format(x) if (x>1) else ''
ppow_fn = lambda x, sep=',': exponents(x,sep=sep) if (x>1) else ''
vpow_fn = lambda x: 'p{}'.format(x) if (x>1) else ''
lpow_fn = lambda x: '^<LBRACKET>{}<RBRACKET>'.format(x) if (x>1) else ''

# subcripts formatting functions
bsub_fn = lambda x: '_{}'.format(x) if (x is not None) else ''
psub_fn = lambda x, sep=',': subscripts(x,sep=sep) if (x is not None) else ''
vsub_fn = lambda x: 's{}'.format(x) if (x is not None) else ''
lsub_fn = lambda x: '_<LBRACKET>{}<RBRACKET>'.format(x) if (x is not None) else ''

# components formatting functions
bcomp_fn = lambda x: ','.join(to_str(x)) if (x is not None) else ''
pcomp_fn = lambda x, sep=',': subscripts(x,sep=sep) if (x is not None) else ''
vcomp_fn = lambda x: '_'+'_'.join(to_str(x)) if (x is not None) else ''
lcomp_fn = lambda x: '_<LBRACKET>{}<RBRACKET>'.format(','.join(to_str(x))) if (x is not None) else ''

# join formatting functions
bjoin_fn = lambda x: '_'.join(to_str(x)) if (x is not None) else ''
pjoin_fn = lambda x: ''.join(to_str(x)) if (x is not None) else ''
vjoin_fn = lambda x: '_'.join(to_str(x)) if (x is not None) else ''
ljoin_fn = lambda x: ''.join(to_str(x)) if (x is not None) else ''

# divide formatting functions
bdivide_fn = lambda x,y: '{}/{}'.format(x,y)
pdivide_fn = lambda x,y: '{}/{}'.format(*to_str(x,y))
vdivide_fn = lambda x,y: '{}__{}'.format(x,y)
ldivide_fn = lambda x,y: r'\dfrac<LBRACKET>{}<RBRACKET><LBRACKET>{}<RBRACKET>'.format(x,y)


class DifferentialStringFormatter(object):
    """
    Utility class to format differential related strings like partial derivatives.

    All string formatting function returns 4 different results:
        *A string that can be used as identifier (name).
        *A pretty string in utf-8 (pretty_name).
        *A variable name that can be used as a valid C identifier for code generation (var_name).
        *A latex string that can be compiled and displayed with latex (latex_name).

    Prefix used for methods:
        b = name
        p = pretty_name
        v = var_name
        l = latex_name

    See __main__ at the bottom of this file for usage.
    """

    exp_fns    = (bexp_fn,  pexp_fn,  vexp_fn,  lexp_fn)
    pow_fns    = (bpow_fn,  ppow_fn,  vpow_fn,  lpow_fn)
    sub_fns    = (bsub_fn,  psub_fn,  vsub_fn,  lsub_fn)
    comp_fns   = (bcomp_fn, pcomp_fn, vcomp_fn, lcomp_fn)
    join_fns   = (bjoin_fn, pjoin_fn, vjoin_fn, ljoin_fn)
    divide_fns = (bdivide_fn, pdivide_fn, vdivide_fn, ldivide_fn)

    @staticmethod
    def format_special_characters(ss):
        special_characters = {
            '<LBRACKET>': '{',
            '<RBRACKET>': '}',
        }
        for (k,v) in special_characters.items():
            ss = ss.replace(k,v)
        return ss

    @classmethod
    def return_names(cls, *args, **kwds):
        # fsc = format special characters
        fsc=kwds.get('fsc', True)
        assert len(args)>=1
        if len(args)==1:
            if fsc:
                return args[0]
            else:
                cls.format_special_characters(args[0])
        else:
            if fsc:
                return tuple(cls.format_special_characters(a) for a in args)
            else:
                return args

    @classmethod
    def format_partial_name(cls, bvar, pvar, vvar, lvar,
                            bpow_fn=bpow_fn, ppow_fn=ppow_fn, vpow_fn=vpow_fn, lpow_fn=lpow_fn,
                            bcomp_fn=bcomp_fn, pcomp_fn=pcomp_fn, vcomp_fn=vcomp_fn, lcomp_fn=lcomp_fn,
                            blp='(', plp='', vlp='', llp='',
                            brp=')', prp='', vrp='',  lrp='',
                            bd='d', pd=partial, vd='d', ld=r'<LBRACKET>\partial<RBRACKET>',
                            dpow=1, varpow=1, components=None,
                            trigp=3, fsc=True):
        assert (varpow != 0)
        bd  = '' if (dpow==0) else bd
        pd  = '' if (dpow==0) else pd
        vd  = '' if (dpow==0) else vd
        ld  = '' if (dpow==0) else ld
        blp = '' if len(bvar) <= trigp else blp
        brp = '' if len(bvar) <= trigp else brp
        plp = '' if len(pvar) <= trigp else plp
        prp = '' if len(pvar) <= trigp else prp
        vlp = '' if len(vvar) <= trigp else vlp
        vrp = '' if len(vvar) <= trigp else vrp
        llp = '' if len(lvar) <= trigp else llp
        lrp = '' if len(lvar) <= trigp else lrp
        template='{d}{dpow}{lp}{var}{components}{rp}{varpow}'
        bname = template.format(d=bd, dpow=bpow_fn(dpow),
                                components=bcomp_fn(components),
                                var=bvar, varpow=bpow_fn(varpow),
                                lp=blp, rp=brp)
        pname = template.format(d=pd, dpow=ppow_fn(dpow),
                                components=pcomp_fn(components),
                                var=pvar, varpow=ppow_fn(varpow),
                                lp=plp, rp=prp)
        vname = template.format(d=vd, dpow=vpow_fn(dpow),
                                components=vcomp_fn(components),
                                var=vvar, varpow=vpow_fn(varpow),
                                lp=vlp, rp=vrp)
        lname = template.format(d=ld, dpow=lpow_fn(dpow),
                                components=lcomp_fn(components),
                                var=lvar, varpow=lpow_fn(varpow),
                                lp=llp, rp=lrp)
        return cls.return_names(bname, pname, vname, lname, fsc=fsc)

    @classmethod
    def format_partial_names(cls, bvars, pvars, vvars, lvars, varpows,
                            bjoin_fn=bjoin_fn, pjoin_fn=pjoin_fn, vjoin_fn=vjoin_fn, ljoin_fn=ljoin_fn,
                            components=None, fsc=True, **kwds):
        bvars, pvars, vvars, lvars = to_tuple(bvars), to_tuple(pvars), to_tuple(vvars), to_tuple(lvars)
        varpows = to_tuple(varpows)
        assert len(bvars)==len(pvars)==len(vvars)==len(lvars)==len(varpows)
        assert any(v>0 for v in varpows)
        nvars = len(bvars)
        if (components is not None):
            components = to_tuple(components)
            assert len(components)==nvars
        else:
            components = (None,)*nvars

        bnames, pnames, vnames, lnames = (), (), (), ()
        for (bvar, pvar, vvar, lvar, varpow, component) in \
                zip(bvars, pvars, vvars, lvars, varpows, components):
            if (varpow==0):
                continue
            res = cls.format_partial_name(bvar=bvar, pvar=pvar, vvar=vvar, lvar=lvar,
                                          varpow=varpow, components=component,
                                          fsc=False, **kwds)
            assert len(res)==4
            bnames += (res[0],)
            pnames += (res[1],)
            vnames += (res[2],)
            lnames += (res[3],)
        return cls.return_names(bjoin_fn(bnames), pjoin_fn(pnames),
                                vjoin_fn(vnames), ljoin_fn(lnames), fsc=fsc)

    @classmethod
    def format_pd(cls, bvar, pvar, vvar, lvar,
                       bxvars='x', pxvars=xsymbol, vxvars='x', lxvars='x',
                       varpows=1, var_components=None, xvars_components=None,
                       bdivide_fn=bdivide_fn, pdivide_fn=pdivide_fn, vdivide_fn=vdivide_fn, ldivide_fn=ldivide_fn,
                       fsc=True, **kwds):

        for k in ('dpow', 'components', 'bvars', 'pvars', 'vvars', 'lvars', 'varpow'):
            assert k not in kwds, 'Cannot specify reserved keyword {}.'.format(k)

        bxvars, pxvars, vxvars, lxvars = to_tuple(bxvars), to_tuple(pxvars), to_tuple(vxvars), to_tuple(lxvars)
        varpows = to_tuple(varpows)
        assert len(bxvars)==len(pxvars)==len(vxvars)==len(lxvars)==len(varpows)
        assert any(v>0 for v in varpows)
        dpow = sum(varpows)

        numerator = cls.format_partial_name(bvar=bvar, pvar=pvar,
                                            vvar=vvar, lvar=lvar,
                                            fsc=False, dpow=dpow,
                                            components=var_components,
                                            **kwds)

        denominator = cls.format_partial_names(bvars=bxvars, pvars=pxvars,
                                               vvars=vxvars, lvars=lxvars,
                                               fsc=False, varpows=varpows,
                                               components=xvars_components,
                                               **kwds)

        return cls.return_names(bdivide_fn(numerator[0], denominator[0]),
                                pdivide_fn(numerator[1], denominator[1]),
                                vdivide_fn(numerator[2], denominator[2]),
                                ldivide_fn(numerator[3], denominator[3]), fsc=fsc)


if __name__ == '__main__':
    def _print(*args, **kwds):
        if isinstance(args[0], tuple):
            assert len(args)==1
            args = args[0]
        if ('multiline' in kwds) and (kwds['multiline'] is True):
            for a in args:
                print(a)
        else:
            print(', '.join(a for a in args))

    print
    bvar, pvar, vvar, lvar = 'Fext', 'Fₑₓₜ', 'Fext', '<LBRACKET>F_<LBRACKET>ext<RBRACKET><RBRACKET>'
    _print(DifferentialStringFormatter.return_names(bvar, pvar, vvar, lvar))

    print
    _print(DifferentialStringFormatter.format_partial_name(bvar, pvar, vvar, lvar, dpow=0))
    _print(DifferentialStringFormatter.format_partial_name(bvar, pvar, vvar, lvar, dpow=1))
    _print(DifferentialStringFormatter.format_partial_name(bvar, pvar, vvar, lvar, dpow=2))
    _print(DifferentialStringFormatter.format_partial_name(bvar, pvar, vvar, lvar, dpow=3, components=0))
    _print(DifferentialStringFormatter.format_partial_name(bvar, pvar, vvar, lvar, dpow=4, components=(0,2)))

    print
    bvar, pvar, vvar, lvar = ('x',)*4
    _print(DifferentialStringFormatter.format_partial_name(bvar, pvar, vvar, lvar, varpow=1))
    _print(DifferentialStringFormatter.format_partial_name(bvar, pvar, vvar, lvar, varpow=2))
    _print(DifferentialStringFormatter.format_partial_name(bvar, pvar, vvar, lvar, varpow=3, components=0))
    _print(DifferentialStringFormatter.format_partial_name(bvar, pvar, vvar, lvar, varpow=4, components=(0,2)))

    print
    bvar, pvar, vvar, lvar = (('x','y'),)*4
    try:
        _print(DifferentialStringFormatter.format_partial_names(bvar, pvar, vvar, lvar, varpows=(0,0)))
        raise RuntimeError()
    except AssertionError:
        pass
    _print(DifferentialStringFormatter.format_partial_names(bvar, pvar, vvar, lvar, varpows=(0,1)))
    _print(DifferentialStringFormatter.format_partial_names(bvar, pvar, vvar, lvar, varpows=(1,0)))
    _print(DifferentialStringFormatter.format_partial_names(bvar, pvar, vvar, lvar, varpows=(1,1)))
    _print(DifferentialStringFormatter.format_partial_names(bvar, pvar, vvar, lvar, varpows=(1,2)))
    _print(DifferentialStringFormatter.format_partial_names(bvar, pvar, vvar, lvar, varpows=(2,2)))
    _print(DifferentialStringFormatter.format_partial_names(bvar, pvar, vvar, lvar, varpows=(2,2), components=(0,1)))
    _print(DifferentialStringFormatter.format_partial_names(bvar, pvar, vvar, lvar, varpows=(2,2), components=((0,1),(1,0))))

    print
    bvar, pvar, vvar, lvar = 'Fext', 'Fₑₓₜ', 'Fext', '<LBRACKET>F_<LBRACKET>ext<RBRACKET><RBRACKET>'
    bxvars, pxvars, vxvars, lxvars = (('x','y'),)*4
    _print(DifferentialStringFormatter.format_pd(bvar, pvar, vvar, lvar))
    _print(DifferentialStringFormatter.format_pd(bvar, pvar, vvar, lvar, varpows=2))
    _print(DifferentialStringFormatter.format_pd(bvar, pvar, vvar, lvar,
                                              bxvars, pxvars, vxvars, lxvars,
                                              varpows=(1,0)))
    _print(DifferentialStringFormatter.format_pd(bvar, pvar, vvar, lvar,
                                              bxvars, pxvars, vxvars, lxvars,
                                              varpows=(0,1)))
    _print(DifferentialStringFormatter.format_pd(bvar, pvar, vvar, lvar,
                                              bxvars, pxvars, vxvars, lxvars,
                                              varpows=(1,1)))
    _print(DifferentialStringFormatter.format_pd(bvar, pvar, vvar, lvar,
                                              bxvars, pxvars, vxvars, lxvars,
                                              varpows=(5,2)))

    print
    bxvars, pxvars, vxvars, lxvars = (('x',)*5,)*4
    varpows = (1,)*5
    xvars_components = tuple(range(5))
    var_components=(0,4,3,2)
    _print(DifferentialStringFormatter.format_pd(bvar, pvar, vvar, lvar,
                                              bxvars, pxvars, vxvars, lxvars,
                                              varpows=varpows, xvars_components=xvars_components,
                                              var_components=var_components), multiline=True)

