"""
Tools to collect time profiling information
for hysop classes.
"""
from hysop.core.mpi import main_rank
from hysop.tools.numpywrappers import npw
from hysop.tools.units import time2str
import numpy as np


class FProfiler(object):
    """
    Class for time measurments on the fly.
    The objects can be linked to a class method.
    """
    def __init__(self, fname):
        """Object to profile on the fly sections of code, methods ...

        Usage:

        >>> from hysop.tools.profiler import FProfiler, ftime
        >>> prof = FProfiler('some_name')
        >>> start = ftime()
        >>> # do something ... [1]
        >>> prof += ftime() - start
        >>> # do something else ... [2]
        >>> prof += ftime() - start
        >>> # ...
        >>> # print(prof)
        >>> # --> display total time spent in do [1] and [2]
        >>> # and number of calls of prof

        """
        # Function name
        self.fname = fname
        # Total execution time
        self.total_time = 0.
        # Number of calls of the fprofiler
        self.nb_calls = 0

    def get_name(self):
        """Profiler name"""
        return self.fname

    def __iadd__(self, t):
        """+= operator"""
        self.total_time += t
        self.nb_calls += 1
        return self

    def __str__(self):
        if self.nb_calls > 0:
            s = '{} ncalls={}, total={}, mean={}'.format(self.fname,
                    self.nb_calls,
                    time2str(self.total_time),
                    time2str(self.total_time/self.nb_calls))
        else:
            s = ''
        return s


class Profiler(object):
    """
    Object used to collect profiling information inside operators.
    """

    def __init__(self, obj, comm):
        """
        Collect profiling information for all operator
        method decorate with @profile.

        Parameters
        ----------
        obj : object (python class) instance. See requirements in notes below.

        Notes:

        * obj must have '_get_profiling_info' and 'name'
        attribute/method.
        """
        self.summary = {}
        self.table = []
        self._comm = comm
        self._comm_size = comm.Get_size()
        for _ in range(self._comm_size):
            self.table.append([])
        # A dictionnary of profiled functions/methods as keys
        # and elapsed time as value.
        self._elems = {}
        # profiled object
        self._obj = obj

        self._l = 1
        self.all_times = None
        self.all_call_nb = None
        self.all_names = [None]

    def down(self, l):
        self._l = l + 1

    def get_name(self):
        """Return the name of the profiled object
        """
        _name = self._obj.name
        return _name

    def __iadd__(self, other):
        """+= operator. Append a new profiled function to the collection"""
        self._elems[other.get_name()] = other
        return self

    def __setitem__(self, key, value):
        self._elems[key] = value

    def __getitem__(self, item):
        try:
            return self._elems[item]
        except KeyError:
            self._elems[item] = FProfiler(item)
            return self._elems[item]

    def __str__(self):
        summary = self.summary
        if len(summary) > 0:
            if (self._l>1) and (len(summary) == 1) and \
                    isinstance(next(iter(summary.values())), FProfiler):
                s = '>{}::{}'.format(self.get_name(), next(iter(summary.values())))
            else:
                s = '{}[{}]>{}{}'.format(
                    '\n' if (self._l==1) else '',
                    main_rank,
                    self.get_name(),
                    ' profiler report' if (self._l==1) else '')
                for v in summary.values():#sorted(summary.values(), key=lambda x: x.total_time):
                    if len(str(v)) > 0:
                        s += '\n{}'.format('  '*self._l + str(v))
        else:
            s = ''
        return s

    def write(self, prefix='', hprefix='', with_head=True):
        """

        Parameters
        ----------
        prefix : string, optional
        hprefix : string, optional
        with_head : bool, optional

        """
        if prefix != '' and prefix[-1] != ' ':
            prefix += ' '
        if hprefix != '' and hprefix[-1] != ' ':
            hprefix += ' '
        if self._comm.Get_rank() == 0:
            s = ""
            h = hprefix + "Rank"
            for r in range(self._comm_size):
                s += prefix + "{0}".format(r)
                for i in range(len(self.all_names)):
                    s += " {0}".format(self.all_times[i][r])
                s += "\n"
            s += prefix + "-1"
            for i in range(len(self.all_names)):
                h += ' ' + self.all_names[i]
                s += " {0}".format(self.all_times[i][self._comm_size])
            h += "\n"
            if with_head:
                s = h + s
            print(s)

    def summarize(self):
        """
        Update profiling values and prepare data for a report
        with print or write.
        """
        # reset summary
        self.summary = {}

        # collect profiling results from decorated object(s), if any.
        self._obj._get_profiling_info()

        from hysop.fields.continuous_field import Field
        i = 0
        for k in self._elems.keys():
            try:
                # Either elem[k] is a FProfiler ...
                self.summary[self._elems[k].total_time] = self._elems[k]
            except AttributeError:
                # ... or a Profiler
                i += 1
                self._elems[k].down(self._l)
                self._elems[k].summarize()
                if isinstance(self._elems[k]._obj, Field):
                    self.summary[1e10 * i] = self._elems[k]
                else:
                    self.summary[1e8 * i] = self._elems[k]
        rk = self._comm.Get_rank()
        for k in sorted(self._elems.keys()):
            if isinstance(self._elems[k], FProfiler):
                self.table[rk].append(
                    (self.get_name() + '.' + k,
                     self._elems[k].total_time, self._elems[k].nb_calls))
        for k in sorted(self._elems.keys()):
            if isinstance(self._elems[k], Profiler):
                for e in self._elems[k].table[rk]:
                    self.table[rk].append(
                        (self.get_name() + '.' + e[0], e[1], e[2]))

        if self._l == 1 and self.all_times is None:
            nb = len(self.table[rk])
            self.all_times = npw.zeros((nb, self._comm_size + 1), order='F')
            self.all_call_nb = npw.integer_zeros((nb, self._comm_size + 1), order='F')
            self.all_names = [None] * nb
            for i in range(nb):
                self.all_names[i]       = self.table[rk][i][0]
                self.all_times[i, rk]   = self.table[rk][i][1]
                self.all_call_nb[i, rk] = self.table[rk][i][2]
            for p in range(self._comm_size):
                self._comm.Bcast(self.all_times[:, p],   root=p)
                self._comm.Bcast(self.all_call_nb[:, p], root=p)
            self.all_times[:, self._comm_size] = np.sum(
                self.all_times[:, :self._comm_size], axis=1)
            self.all_call_nb[:, self._comm_size] = np.sum(
                self.all_call_nb[:, :self._comm_size], axis=1)
