## HySoP Example: Turbulent scalar advection-diffusion
## Example from JM Etancelin PhD (section 6.3) BUT Without Tasks for current HySoP version

import os
import numpy as np
pi  = np.pi
cos = np.cos
sin = np.sin

width = 0.01
ampl3 = 0.3
ampl = 0.05


## Function to compute initial vorticity
def init_velocity(data, coords, component):
    (x, y, z) = coords
    yy = (0.1 - 2. * np.abs(y - 0.5)) / (4. * width)
    strg = np.exp(-np.abs(yy ** 2)) * np.random.random(size=data.shape)
    if (component == 0):
        data[...] = 0.5 * (1. + np.tanh(yy))
        data[...] *= (1. + ampl3 * np.sin(8. * np.pi * x))
        data[...] *= (1. + ampl * strg)
    else:
        data[...] = ampl * strg
    return data

## Function to compute initial vorticity
def init_vorticity(data, coords, component):
    (x, y, z) = coords
    if (component==2):
        xx = -(0.3*np.sin(8.0*np.pi*x) + 1.0)
        yp=25.0*np.tanh(-50.0*y[y>0.5] + 27.5)**2 - 25.0
        ym=-25.0*np.tanh(50.0*y[y<0.5] - 22.5)**2 + 25.0
        data[:,y[0,:,0]>0.5,:] = yp[np.newaxis,:,np.newaxis]
        data[:,y[0,:,0]==0.5,:] = 0.
        data[:,y[0,:,0]<0.5,:] = ym[np.newaxis,:,np.newaxis]
        data[...] *= xx
    else:
        data[...] = 0.

def init_scal(data, coords, component):
    assert (component == 0)
    (x, y, z) = coords
    yy = abs(y - 0.5)
    aux = (0.1 - 2. * yy) / (4. * width)
    data[...] = 0.5 * (1. + np.tanh(aux))
    data[...] *= (1. + ampl3 * np.sin(8. * pi * x))


from hysop import Box, Simulation, Problem, MPIParams, Field, IOParams
from hysop.defaults import VelocityField, VorticityField, \
                           EnstrophyParameter, TimeParameters
from hysop.constants import Implementation, AdvectionCriteria, HYSOP_REAL, \
    StretchingFormulation
from hysop.topology.cartesian_topology import CartesianTopology
from hysop.operators import Advection, StaticDirectionalStretching, Diffusion, \
                            PoissonCurl, AdaptiveTimeStep, DirectionalDiffusion, \
                            Enstrophy, MinMaxFieldStatistics, StrangSplitting,    \
                            ParameterPlotter, DirectionalAdvection
from hysop.numerics.odesolvers.runge_kutta import RK2
from hysop.methods import SpaceDiscretization, Remesh, TimeIntegrator, \
                          ComputeGranularity, Interpolation, StrangOrder
from hysop.tools.parameters import CartesianDiscretization

# Define the domain
dim  = 3
npts_uw = (64, 64, 128)
npts_s = (128, 128, 256)
box  = Box(origin=(0., )*dim, length=(1.,1., 2.), dim=dim)
# Physical parameters:
# Flow viscosity
VISCOSITY = 1e-4
# Schmidt number
SC = ((1. * npts_s[0] - 1.)/(1. * npts_uw[0] - 1.))**2
# Scalar diffusivity
DIFF_COEFF_SCAL = VISCOSITY / SC
dt_max = 0.012
cfl = 1.5
lcfl = 0.15
dump_freq = 100

# Get default MPI Parameters from domain (even for serial jobs)
mpi_params = MPIParams(comm=box.task_comm,
                       task_id=box.current_task())

# Create an explicit OpenCL context from user parameters
from hysop.backend.device.opencl.opencl_tools import get_or_create_opencl_env, get_device_number
cl_env = get_or_create_opencl_env(
    mpi_params=mpi_params,
    platform_id=0,
    device_id=box.machine_rank%get_device_number())
# Setup usual implementation specific variables
impl = None
extra_op_kwds = {'mpi_params': mpi_params,
                 'cl_env': cl_env}
method = {}

# Define parameters and field (time, timestep, velocity, vorticity, enstrophy)
t, dt = TimeParameters(dtype=HYSOP_REAL)
velo = VelocityField(domain=box, dtype=HYSOP_REAL)
vorti = VorticityField(velocity=velo, dtype=HYSOP_REAL)
enstrophy = EnstrophyParameter(dtype=HYSOP_REAL)
wdotw = Field(domain=box, dtype=HYSOP_REAL, is_vector=False, name="WdotW")
scal = Field(domain=box, name='Scalar', is_vector=False)

# Topologies
topo_nogh = CartesianTopology(domain=box,
                              discretization=CartesianDiscretization(npts_uw,
                                  default_boundaries=True),
                              mpi_params=mpi_params,
                              cutdirs=[True, False, False])

### Build the directional operators
#> Directional advection
advec_scal = DirectionalAdvection(
    implementation=Implementation.OPENCL,
    name='advec_scal',
    velocity=velo,
    velocity_cfl=dt_max*1.2*(npts_uw[0]-1),
    advected_fields=(scal,),
    variables={velo: npts_uw, scal: npts_s},
    dt=dt, **extra_op_kwds)
diffuse_scal = DirectionalDiffusion(
    implementation=Implementation.OPENCL,
    name='diffuse_scal',
    fields=(scal,),
    coeffs=DIFF_COEFF_SCAL,
    variables={scal: npts_s},
    dt=dt, **extra_op_kwds)
splitting_scal = StrangSplitting(splitting_dim=dim,
                            order=StrangOrder.STRANG_FIRST_ORDER)
splitting_scal.push_operators(advec_scal)
splitting_scal.push_operators(diffuse_scal)
#> Directional advection
advec = Advection(
    implementation=Implementation.FORTRAN,
    name='advec',
    velocity=velo,
    advected_fields=(vorti,),
    variables={velo: npts_uw, vorti: npts_uw},
    dt=dt, **extra_op_kwds)
#> Directional stretching
stretch = StaticDirectionalStretching(
    implementation=Implementation.PYTHON,
    name='S',
    formulation=StretchingFormulation.CONSERVATIVE,
    velocity=velo,
    vorticity=vorti,
    variables={velo: npts_uw, vorti: npts_uw},
    dt=dt, **extra_op_kwds)
#> Directional splitting operator subgraph
splitting = StrangSplitting(splitting_dim=dim,
                            order=StrangOrder.STRANG_SECOND_ORDER)
splitting.push_operators(stretch)
#> Diffusion
diffuse = Diffusion(
    implementation=Implementation.FORTRAN,
    name='diffuse',
    nu=VISCOSITY,
    Fin=vorti,
    variables={vorti: topo_nogh},
    dt=dt, **extra_op_kwds)
#> Poisson operator to recover the velocity from the vorticity
poisson = PoissonCurl(
    implementation=Implementation.FORTRAN,
    name='poisson',
    velocity=velo,
    vorticity=vorti,
    variables={velo: topo_nogh, vorti: topo_nogh},
    projection=None,
    **extra_op_kwds)

#> Operator to compute the infinite norm of the velocity
min_max_U = MinMaxFieldStatistics(
    name='min_max_U', field=velo,
    Finf=True, implementation=Implementation.PYTHON,
    variables={velo: npts_uw},
    **extra_op_kwds)
#> Operator to compute the infinite norm of the vorticity
min_max_W = MinMaxFieldStatistics(
    name='min_max_W', field=vorti,
    Finf=True, implementation=Implementation.PYTHON,
    variables={vorti: npts_uw},
    **extra_op_kwds)
#> Operator to compute the enstrophy
enstrophy_op = Enstrophy(
    name='enstrophy', vorticity=vorti,
    enstrophy=enstrophy, WdotW=wdotw,
    variables={vorti: topo_nogh, wdotw: topo_nogh},
    implementation=Implementation.PYTHON, **extra_op_kwds)

### Adaptive timestep operator
adapt_dt = AdaptiveTimeStep(dt, max_dt=dt_max, equivalent_CFL=True)
dt_cfl = adapt_dt.push_cfl_criteria(cfl=cfl, Finf=min_max_U.Finf,
                                    equivalent_CFL=True)
dt_advec = adapt_dt.push_advection_criteria(lcfl=lcfl, Finf=min_max_W.Finf,
                                            criteria=AdvectionCriteria.W_INF)

#> We ask to dump the outputs of this operator
io_params = IOParams(filename='dummy', dump_freq=dump_freq)
poisson.dump_inputs(fields=(vorti,), io_params=io_params.clone(filename='vorti'))
poisson.dump_outputs(fields=(velo,), io_params=io_params.clone(filename='velo'))
splitting_scal.dump_outputs(fields=(scal,), frequency=dump_freq, io_params=io_params.clone(filename='scalar'))

## Create the problem we want to solve and insert our
# directional splitting subgraph and the standard operators.
# The method dictionnary passed to this graph will be dispatched
# accross all operators contained in the graph.
method.update({ComputeGranularity:    0,
               SpaceDiscretization:   4,
               TimeIntegrator:        RK2,
               Remesh:                Remesh.L4_2,
               Interpolation:         Interpolation.LINEAR})
problem = Problem(method=method)
problem.insert(advec, splitting, diffuse, splitting_scal, poisson,
               enstrophy_op, min_max_U, min_max_W, adapt_dt)
problem.build()

# Create a simulation
# (do not forget to specify the t and dt parameters here)
simu = Simulation(start=0., end=5.,
                  max_iter=9999,
                  dt0=1e-5,
                  times_of_interest=(4.5, ),
                  t=t, dt=dt)
simu.write_parameters(t, dt_cfl, dt_advec, dt, enstrophy,
                      min_max_U.Finf, min_max_W.Finf, adapt_dt.equivalent_CFL,
                      filename='parameters.txt', precision=8)

# Initialize only the vorticity
problem.initialize_field(velo, formula=init_velocity)
problem.initialize_field(vorti, formula=init_vorticity)
problem.initialize_field(scal, formula=init_scal)

# Finally solve the problem
problem.solve(simu)

# Finalize
problem.finalize()
