import numpy as np
import scipy as sp
import sympy as sm
import numba as nb
import skfmm

TANK_RATIO      = 1
FILL_PCT        = 0.25
SEDIMENT_RADIUS = 5e-3
SEDIMENT_COUNT  = int(1.15*(FILL_PCT*TANK_RATIO / (np.pi*SEDIMENT_RADIUS**2)))
DISCRETIZATION  = 256

BLOB_INIT = False
NB = 5

# initialize vorticity
def init_vorticity(data, coords, component=None):
    # the flow is initially quiescent
    for d in data:
        d[...] = 0.0

# initialize velocity
def init_velocity(data, coords, component=None):
    # the flow is initially quiescent
    for d in data:
        d[...] = 0.0

def init_phi(data, coords, nblobs, rblob, component):
    from hysop import vprint
    X, Y = coords[0], coords[1]
    Bx = np.random.rand(nblobs)
    By = TANK_RATIO*np.random.rand(nblobs)
    R2 = rblob * rblob

    cache_file='/tmp/C_init_ls_{}_{}'.format('_'.join(str(x) for x in data.shape),
            str(abs(hash((BLOB_INIT, NB, TANK_RATIO, nblobs, rblob)))))
    try:
        D = np.load(file=cache_file+'.npz')
        vprint('  *Initializing sediments from cache: "{}.npz".'.format(cache_file))
        data[...] = D['data']
    except:

        if not BLOB_INIT:
            vprint('  *Initializing sediments with sines...')
            data[...] = np.sin(2*np.pi*NB*X)*np.sin(2*np.pi*NB*Y)
            data[...] = np.abs(data)
            return

        # we cache initialization
        vprint('  *Initializing sediments of radius {} with {} random blobs.'.format(rblob, nblobs))
        np.savez_compressed(file=cache_file, data=data)
        vprint('  *Caching data to "{}.npz".'.format(cache_file))

        X,   Y   = X.ravel(), Y.ravel()
        dx,  dy  = X[1]-X[0], Y[1]-Y[0]
        Nx,  Ny  = X.size, Y.size
        Rx,  Ry  = 2*(int(rblob/dx)+1), 2*(int(rblob/dy)+1)
        assert (rblob>=dx), 'Sediment radius < dx.'
        assert (rblob>=dy), 'Sediment radius < dy.'

        Ix = np.floor(Bx/dx).astype(np.int32)
        Iy = np.floor(By/dy).astype(np.int32)
        Px = Bx - Ix*dx
        Py = By - Iy*dy

        from hysop.tools.numba_utils import make_numba_signature
        args = (Ix, Iy, Bx, By, data)
        signature, _ = make_numba_signature(*args)

        @nb.guvectorize([signature],
            '(n),(n),(n),(n),(n0,n1)',
            target='parallel',
            nopython=True, cache=True)
        def iter_blobs(Ix, Iy, Bx, By, data):
            for k in range(nblobs):
                ix, iy = Ix[k], Iy[k]
                px, py = Px[k], Py[k]
                for i in range(-Ry, +Ry):
                    ii = iy+i
                    if (ii<0) or (ii>=Ny):
                        continue
                    dy2 = (py + i*dy)**2 / R2
                    for j in range(-Rx, +Rx):
                        jj = ix+j
                        if (jj<0) or (jj>=Nx):
                            continue
                        dx2 = (px - j*dx)**2 / R2
                        d = dx2 + dy2 - 1
                        if (d<data[ii,jj]):
                            data[ii,jj] = d

        vprint('  *Initializing sediments of radius {} with {} random blobs.'.format(rblob, nblobs))
        data[...]  = np.inf
        iter_blobs(*args)
        data[...] = skfmm.distance(data, dx=(dy,dx))

        # we cache initialization
        np.savez_compressed(file=cache_file, data=data)
        vprint('  *Caching data to "{}.npz".'.format(cache_file))



def compute(args):
    from hysop import Field, Box, Simulation, Problem, MPIParams, IOParams, vprint, \
                      ScalarParameter
    from hysop.defaults import VelocityField, VorticityField, \
                               DensityField, ViscosityField, \
                               LevelSetField, PenalizationField, \
                               EnstrophyParameter, TimeParameters, \
                               VolumicIntegrationParameter
    from hysop.constants import Implementation, AdvectionCriteria, \
                                BoxBoundaryCondition, BoundaryCondition, \
                                Backend

    from hysop.operators import DirectionalAdvection, DirectionalStretching,       \
                                Diffusion, ComputeMeanField,                       \
                                PoissonCurl, AdaptiveTimeStep,                     \
                                Enstrophy, MinMaxFieldStatistics, StrangSplitting, \
                                ParameterPlotter, Integrate, HDF_Writer,           \
                                CustomSymbolicOperator, DirectionalSymbolic,       \
                                SpectralExternalForce, SymbolicExternalForce

    from hysop.methods import SpaceDiscretization, Remesh, TimeIntegrator, \
                              ComputeGranularity, Interpolation

    from hysop.numerics.odesolvers.runge_kutta import Euler, RK2, RK3, RK4
    from hysop.symbolic import sm, space_symbols, local_indices_symbols
    from hysop.symbolic.base import SymbolicTensor
    from hysop.symbolic.field import curl
    from hysop.symbolic.relational import Assignment, LogicalLE, LogicalGE, LogicalLT, LogicalGT
    from hysop.symbolic.misc import Select
    from hysop.symbolic.tmp import TmpScalar
    from hysop.tools.string_utils import framed_str

    # Constants
    dim = args.ndim
    if (dim==2):
        Xo = (0.0,)*dim
        Xn = (float(TANK_RATIO), 1.0)
        nblobs = SEDIMENT_COUNT
        rblob  = SEDIMENT_RADIUS
        npts = args.npts
    else:
        msg='The {}D has not been implemented yet.'.format(dim)
        raise NotImplementedError(msg)

    nu_W = ScalarParameter(name='nu_W', dtype=args.dtype, const=True, initial_value=1e-2)

    lboundaries = (BoxBoundaryCondition.SYMMETRIC, BoxBoundaryCondition.SYMMETRIC)
    rboundaries = (BoxBoundaryCondition.SYMMETRIC, BoxBoundaryCondition.SYMMETRIC)

    S_boundaries = {
        'lboundaries': (BoundaryCondition.HOMOGENEOUS_NEUMANN, BoundaryCondition.HOMOGENEOUS_NEUMANN),
        'rboundaries': (BoundaryCondition.HOMOGENEOUS_NEUMANN, BoundaryCondition.HOMOGENEOUS_NEUMANN)
    }

    box = Box(origin=Xo, length=np.subtract(Xn,Xo),
                lboundaries=lboundaries, rboundaries=rboundaries)

    # Get default MPI Parameters from domain (even for serial jobs)
    mpi_params = MPIParams(comm=box.task_comm,
                           task_id=box.current_task())

    # Setup usual implementation specific variables
    impl = args.impl
    enforce_implementation = args.enforce_implementation
    extra_op_kwds = {'mpi_params': mpi_params}
    if (impl is Implementation.PYTHON):
        method = {}
    elif (impl is Implementation.OPENCL):
        # For the OpenCL implementation we need to setup the compute device
        # and configure how the code is generated and compiled at runtime.

        # Create an explicit OpenCL context from user parameters
        from hysop.backend.device.opencl.opencl_tools import get_or_create_opencl_env, get_device_number
        cl_env = get_or_create_opencl_env(
            mpi_params=mpi_params,
            platform_id=args.cl_platform_id,
            device_id=box.machine_rank%get_device_number() if args.cl_device_id is None else None)

        # Configure OpenCL kernel generation and tuning (already done by HysopArgParser)
        from hysop.methods import OpenClKernelConfig
        method = { OpenClKernelConfig: args.opencl_kernel_config }

        # Setup opencl specific extra operator keyword arguments
        extra_op_kwds['cl_env'] = cl_env
    else:
        msg='Unknown implementation \'{}\'.'.format(impl)
        raise ValueError(msg)

    # Define parameters and field (time, timestep, velocity, vorticity, enstrophy)
    t, dt = TimeParameters(dtype=args.dtype)
    velo  = VelocityField(domain=box, dtype=args.dtype)
    vorti = VorticityField(velocity=velo)
    phi   = LevelSetField(domain=box, dtype=args.dtype, **S_boundaries)
    S     = DensityField(name='S', domain=box, dtype=args.dtype, **S_boundaries)
    Sv    = VolumicIntegrationParameter(field=S)

    # Symbolic fields
    frame = velo.domain.frame
    Us    = velo.s(*frame.vars)
    Ws    = vorti.s(*frame.vars)
    phis  = phi.s(*frame.vars)
    Ss  = S.s(*frame.vars)
    dts   = dt.s

    ### Build the directional operators
    #> Directional advection
    advec = DirectionalAdvection(implementation=impl,
            name='advec',
            velocity = velo,
            advected_fields = (vorti, phi),
            velocity_cfl = args.cfl,
            variables = {velo:  npts,
                         vorti: npts,
                         phi:   npts},
            dt=dt, **extra_op_kwds)

    #> Recompute density from levelset
    dx = np.max(np.divide(box.length, np.asarray(args.npts)-1))
    S1, S2 = 0.5, 0.0
    pi  = TmpScalar(name='pi',  dtype=args.dtype)
    eps = TmpScalar(name='eps', dtype=args.dtype)
    x   = TmpScalar(name='x',   dtype=args.dtype)
    H   = TmpScalar(name='H',   dtype=args.dtype)
    smooth_cond = LogicalLT(sm.Abs(x), eps)
    pos_cond = LogicalGT(x, 0)
    clamp  = Select(0.0, 1.0, pos_cond)
    smooth = (x+eps)/(2*eps) + sm.sin(pi*x/eps)/(2*pi)
    H_eps = Select(clamp, smooth, smooth_cond)
    #e0 = Assignment(pi, np.pi)
    #e1 = Assignment(eps, 5*dx)
    #e2 = Assignment(x, phis*SEDIMENT_RADIUS)
    #e3 = Assignment(H, H_eps)
    #e4 = Assignment(Ss, S1 + (S2-S1)*H)
    #exprs = (e0,e1,e2,e3,e4)
    if BLOB_INIT:
        e = Assignment(Ss, 0.5*LogicalLE(phis, 0))
    else:
        e = Assignment(Ss, 0.5*LogicalGT(phis, 0.5))
        #e = Assignment(Ss, 0.5*LogicalLE(phis, 0))
    exprs = (e,)
    eval_fields = DirectionalSymbolic(name='eval_fields',
                                    pretty_name='{}({})'.format(
                                        phi.pretty_name,
                                        S.pretty_name),
                                    no_split=True,
                                    implementation=impl,
                                    exprs=exprs, dt=dt,
                                    variables={phi: npts,
                                               S: npts},
                                    **extra_op_kwds)

    #> Stretch vorticity
    if (dim==3):
        stretch = DirectionalStretching(implementation=impl,
                 name='S',
                 pretty_name='S',
                 formulation = args.stretching_formulation,
                 velocity  = velo,
                 vorticity = vorti,
                 variables = {velo: npts, vorti: npts},
                 dt=dt, **extra_op_kwds)
    elif (dim==2):
        stretch = None
    else:
        msg='Unsupported dimension {}.'.format(dim)
        raise RuntimeError(msg)

    #> External force rot(-S*g)
    Fext = np.zeros(shape=(dim,), dtype=object).view(SymbolicTensor)
    fext = -Ss
    Fext[1] = fext
    lhs = Ws.diff(frame.time)
    rhs = curl(Fext, frame)
    exprs = Assignment.assign(lhs, rhs)
    external_force = DirectionalSymbolic(name='Fext',
                                    implementation=impl,
                                    exprs=exprs, dt=dt,
                                    variables={vorti: npts,
                                               S: npts},
                                    **extra_op_kwds)

    splitting = StrangSplitting(splitting_dim=dim,
                    order=args.strang_order)
    splitting.push_operators(advec, eval_fields, stretch, external_force)

    ### Build standard operators
    #> Poisson operator to recover the velocity from the vorticity
    poisson = PoissonCurl(name='poisson', velocity=velo, vorticity=vorti,
                            variables={velo:npts, vorti: npts},
                            diffusion=nu_W, dt=dt,
                            implementation=impl,
                            enforce_implementation=enforce_implementation,
                            **extra_op_kwds)

    #> Operator to compute the infinite norm of the velocity
    min_max_U = MinMaxFieldStatistics(name='min_max_U', field=velo,
            Finf=True, implementation=impl, variables={velo:npts},
            **extra_op_kwds)
    #> Operator to compute the infinite norm of the vorticity
    min_max_W = MinMaxFieldStatistics(field=vorti,
            Finf=True, implementation=impl, variables={vorti:npts},
            **extra_op_kwds)

    #> Operators to compute the integrated density
    integrate_S = Integrate(field=S, variables={S: npts},
                                    parameter=Sv, scaling='volumic', cst=2,
                                    implementation=impl, **extra_op_kwds)

    #> Operators to dump all fields
    io_params = IOParams(filename='fields', frequency=args.dump_freq)
    dump_fields = HDF_Writer(name='dump',
                             io_params=io_params,
                             force_backend=Backend.OPENCL,
                             variables={#velo: npts,
                                        #vorti: npts,
                                        phi: npts,
                                        S: npts},
                             **extra_op_kwds)

    #> Operator to compute and save mean fields
    axes = list(range(1, dim))
    view = [slice(None,None,None),]*dim
    view = tuple(view)
    io_params = IOParams(filename='horizontally_averaged_profiles', frequency=0)
    compute_mean_fields = ComputeMeanField(name='mean',
            fields={S: (view, axes)},
            variables={S: npts},
            io_params=io_params)

    ### Adaptive timestep operator
    adapt_dt = AdaptiveTimeStep(dt, equivalent_CFL=True,
                                    name='merge_dt', pretty_name='dt',
                                    max_dt=1e-1)
    dt_cfl = adapt_dt.push_cfl_criteria(cfl=args.cfl,
                                        Finf=min_max_U.Finf,
                                        equivalent_CFL=True,
                                        name='dt_cfl', pretty_name='CFL')
    dt_advec = adapt_dt.push_advection_criteria(lcfl=args.lcfl, Finf=min_max_W.Finf,
                                                 criteria=AdvectionCriteria.W_INF,
                                                 name='dt_lcfl', pretty_name='LCFL')


    ## Create the problem we want to solve and insert our
    # directional splitting subgraph and the standard operators.
    # The method dictionnary passed to this graph will be dispatched
    # accross all operators contained in the graph.
    method.update(
            {
               ComputeGranularity:    args.compute_granularity,
               SpaceDiscretization:   args.fd_order,
               TimeIntegrator:        args.time_integrator,
               Remesh:                args.remesh_kernel,
            }
        )

    problem = Problem(method=method)
    problem.insert(poisson,
                   min_max_U, min_max_W, adapt_dt,
                   splitting,
                   integrate_S,
                   dump_fields,
                   compute_mean_fields)
    problem.build(args)

    # If a visu_rank was provided, and show_graph was set,
    # display the graph on the given process rank.
    if args.display_graph:
        problem.display(args.visu_rank)

    # Create a simulation
    # (do not forget to specify the t and dt parameters here)
    simu = Simulation(start=args.tstart, end=args.tend,
                      nb_iter=args.nb_iter,
                      max_iter=args.max_iter,
                      dt0=args.dt, times_of_interest=args.times_of_interest,
                      t=t, dt=dt)
    simu.write_parameters(t, dt_cfl, dt_advec, dt,
            min_max_U.Finf, min_max_W.Finf, adapt_dt.equivalent_CFL,
            filename='parameters.txt', precision=8)

    # Initialize vorticity, velocity, S on all topologies
    problem.initialize_field(field=velo,  formula=init_velocity)
    problem.initialize_field(field=vorti, formula=init_vorticity)
    problem.initialize_field(field=phi,   formula=init_phi, nblobs=nblobs, rblob=rblob,
            without_ghosts=BLOB_INIT)

    # Finally solve the problem
    problem.solve(simu, dry_run=args.dry_run,
            debug_dumper=args.debug_dumper,
            checkpoint_handler=args.checkpoint_handler)


    # Finalize
    problem.finalize()


if __name__=='__main__':
    from hysop_examples.argparser import HysopArgParser, colors

    class ParticleAboveSaltArgParser(HysopArgParser):
        def __init__(self):
            prog_name = 'sediment_deposit_levelset'
            default_dump_dir = '{}/hysop_examples/{}'.format(HysopArgParser.tmp_dir(),
                    prog_name)

            description=colors.color('HySoP Sediment Deposit Levelset Example: ', fg='blue',
                    style='bold')
            description+='\n'
            description+='\nThis example focuses on a validation study for the '
            description+='hybrid particle-mesh vortex method for sediment deposit '
            description+='using the levelset method.'

            super(ParticleAboveSaltArgParser, self).__init__(
                 prog_name=prog_name,
                 description=description,
                 default_dump_dir=default_dump_dir)

        def _setup_parameters(self, args):
            super(ParticleAboveSaltArgParser, self)._setup_parameters(args)
            dim = args.ndim
            if (dim not in (2,3)):
                msg='Domain should be 2D or 3D.'
                self.error(msg)


    parser = ParticleAboveSaltArgParser()

    parser.set_defaults(impl='cl', ndim=2,
                        npts=(TANK_RATIO*DISCRETIZATION+1,DISCRETIZATION+1),
                        box_origin=(0.0,), box_length=(1.0,),
                        tstart=0.0, tend=100.1,
                        dt=1e-6, cfl=0.50, lcfl=0.50,
                        dump_times=tuple(float(1*x) for x in range(100)),
                        dump_freq=0)

    parser.run(compute)

